from .export_mixin import CeleryExportAdminMixin
from .import_mixin import CeleryImportAdminMixin


class CeleryImportExportMixin(
    CeleryImportAdminMixin,
    CeleryExportAdminMixin,
):
    """Import and export mixin."""

    # template for change_list view
    import_export_change_list_template = "admin/import_export/change_list_import_export.html"



import os
import pathlib
import traceback
import uuid
from typing import Optional, Sequence, Type

from django.conf import settings
from django.db import models, transaction
from django.utils import encoding, module_loading, timezone
from django.utils.translation import gettext_lazy as _

import tablib
from celery import current_app, result, states
from import_export.formats import base_formats
from import_export.results import Result
from picklefield.fields import PickledObjectField

from . import tools
from .core import TaskStateInfo, TimeStampedModel


class ImportJob(TimeStampedModel):
    """Abstract model for managing celery import jobs.

    Encapsulate all logic related to celery import.

    Import steps:

    1. Create ``ImportJob`` with resource initialization parameters and file
        with data to be imported from
    2. Dry run. Try to import all data from file and collect statistics (
        errors, new rows, updated rows).
    3. If data for import is correct - import data from file to database.

    """

    class ImportStatus(models.TextChoices):
        """ImportJob possible statuses.

        * CREATED:
            import job is just created, no parsing done
        * PARSING:
            parse job started
        * PARSED:
            data_file parsed, no errors in data occurred
        * INPUT_ERROR:
            data_file parsed, data contain errors
        * PARSE_ERROR:
            data_file can't be parsed (invalid format, etc.)
        * IMPORT_CONFIRMED
            import confirmed but not started yet
        * IMPORTING:
            importing job started
        * IMPORTED:
            data from data_file imported to DB w/o errors
        * IMPORT_ERROR:
            unknown error during import
        * CANCELLED:
            import job has been cancelled (revoked)

        State diagrams::

            CREATED
               |
            .parse_data()
               |
            PARSING  - (INPUT_ERROR, PARSE_ERROR)
               |
            PARSED
               |
            .confirm_import()
               |
            IMPORT_CONFIRMED
               |
            .import_data()
               |
            IMPORTING - IMPORT_ERROR
               |
            IMPORTED

        """

        CREATED = "CREATED", _("Created")
        PARSING = "PARSING", _("Parsing")
        PARSED = "PARSED", _("Parsed")
        INPUT_ERROR = "INPUT_ERROR", _("Input data error")
        PARSE_ERROR = "PARSE_ERROR", _("Parse error")
        CONFIRMED = "CONFIRMED", _("Import confirmed")
        IMPORTING = "IMPORTING", _("Importing")
        IMPORTED = "IMPORTED", _("Imported")
        IMPORT_ERROR = "IMPORT_ERROR", _("Import error")
        CANCELLED = "CANCELLED", _("Cancelled")

    results_statuses = (
        ImportStatus.PARSED,
        ImportStatus.INPUT_ERROR,
        ImportStatus.IMPORTED,
        ImportStatus.IMPORT_ERROR,
    )

    progress_statuses = (
        ImportStatus.PARSING,
        ImportStatus.IMPORTING,
    )

    parse_finished_statuses = (
        ImportStatus.INPUT_ERROR,
        ImportStatus.PARSE_ERROR,
        ImportStatus.PARSED,
    )

    import_finished_statuses = (
        ImportStatus.IMPORTED,
        ImportStatus.IMPORT_ERROR,
    )

    success_statuses = (
        ImportStatus.IMPORTED,
        ImportStatus.PARSED,
    )

    failure_statuses = (
        ImportStatus.INPUT_ERROR,
        ImportStatus.PARSE_ERROR,
    )

    import_status = models.CharField(
        max_length=20,
        choices=ImportStatus.choices,
        default=ImportStatus.CREATED,
        verbose_name=_("Job status"),
    )

    resource_path = models.CharField(
        max_length=128,
        verbose_name=_("Resource class path"),
        help_text=_(
            "Dotted path to subclass of `import_export.Resource` that "
            "should be used for import",
        ),
    )

    data_file = models.FileField(
        max_length=512,
        verbose_name=_("Data file"),
        upload_to=tools.upload_import_file_to,
        help_text=_("File that contain data to be imported"),
    )

    resource_kwargs = models.JSONField(
        default=dict,
        verbose_name=_("Resource kwargs"),
        help_text=_("Keyword parameters required for resource initialization"),
    )

    traceback = models.TextField(
        blank=True,
        default=str,
        verbose_name=_("Traceback"),
        help_text=_("Python traceback in case of parse/import error"),
    )
    error_message = models.CharField(
        max_length=128,
        blank=True,
        default=str,
        verbose_name=_("Error message"),
        help_text=_("Python error message in case of parse/import error"),
    )

    result = PickledObjectField(
        default=str,
        verbose_name=_("Import result"),
        help_text=_(
            "Internal import result object that contain "
            "info about import statistics. Pickled Python object",
        ),
    )

    parse_task_id = models.CharField(
        default=str,
        max_length=36,
        verbose_name=_("Parsing task ID"),
        help_text=_("Celery task ID that start `parse_data`"),
    )

    import_task_id = models.CharField(
        max_length=36,
        default=str,
        verbose_name=_("Import task ID"),
        help_text=_("Celery task ID that start `import_data`"),
    )

    parse_finished = models.DateTimeField(
        editable=False,
        null=True,
        verbose_name=_("Parse finished"),
    )

    import_started = models.DateTimeField(
        editable=False,
        null=True,
        verbose_name=_("Import started"),
    )

    import_finished = models.DateTimeField(
        editable=False,
        null=True,
        verbose_name=_("Import finished"),
    )

    created_by = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Created by"),
        help_text=_("User which started import"),
    )

    skip_parse_step = models.BooleanField(
        default=False,
        help_text=_("Start importing without confirmation"),
        verbose_name=_("Skip parse step"),
    )

    class Meta:
        verbose_name = _("Import job")
        verbose_name_plural = _("Import jobs")

    def __str__(self) -> str:
        """Return string representation."""
        resource_name = pathlib.Path(self.resource_path).suffix.lstrip(".")

        return f"ImportJob(resource={resource_name})"

    def save(
        self,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None,
    ):
        """Start task for data parsing when ImportJob is created.

        Celery task is manually called with `apply_async`, to provide
        possibility of custom `task_id` with which task will be run.

        """
        is_created = self._state.adding
        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )
        if not is_created:
            return

        if self.skip_parse_step:
            self.import_task_id = str(uuid.uuid4())
            self.import_started = timezone.now()
            self.save(
                update_fields=[
                    "import_task_id",
                    "import_started",
                ],
            )
            transaction.on_commit(self._start_import_data_task)
        else:
            self.parse_task_id = str(uuid.uuid4())
            self.save(update_fields=["parse_task_id"])
            transaction.on_commit(self.start_parse_data_task)

    @property
    def resource(self):
        """Get initialized resource instance."""
        resource_class = module_loading.import_string(self.resource_path)
        resource = resource_class(
            **self.resource_kwargs,
        )
        return resource

    @property
    def progress(self) -> Optional[TaskStateInfo]:
        """Return dict with parsing state.

        Example for sync mode::

            {
                'state': 'PARSING',
                'info': None
            }

        Example for background (celery) mode::

            {
                'state': 'PARSING',
                'info': {'current': 15, 'total': 100}
            }

        Possible states:
            1. PENDING
            2. STARTED
            3. SUCCESS
            4. PARSING - custom status that also set importing info

        https://docs.celeryproject.org/en/latest/userguide/tasks.html#states

        """
        if self.import_status not in (
            self.ImportStatus.PARSING,
            self.ImportStatus.IMPORTING,
        ):
            return None

        current_task = (
            self.parse_task_id
            if self.import_status == self.ImportStatus.PARSING
            else self.import_task_id
        )

        if not current_task or current_app.conf.task_always_eager:
            return dict(
                state=self.import_status.upper(),
                info=None,
            )

        return self._get_task_state(current_task)

    def _check_import_status_correctness(
        self,
        expected_statuses: Sequence[ImportStatus],
    ) -> None:
        """Raise `ValueError` if `ImportJob` is in incorrect state."""
        if self.import_status not in expected_statuses:
            raise ValueError(
                f"ImportJob with id {self.id} has incorrect status: "
                f"`{self.import_status}`. Expected statuses:"
                f" {[status.value for status in expected_statuses]}",
            )

    def start_parse_data_task(self):
        """Start parsing task."""
        from .. import tasks

        tasks.parse_data_task.apply_async(
            kwargs=dict(job_id=self.pk),
            task_id=self.parse_task_id,
        )

    def parse_data(self):
        """Parse `data_file` and collect results.

        Sets `result` and/or `traceback` and update `status`.

        """
        self._check_import_status_correctness(
            expected_statuses=(self.ImportStatus.CREATED,),
        )

        self.import_status = self.ImportStatus.PARSING
        self.save(update_fields=["import_status"])

        try:
            self.result = self._parse_data_inner()
            self.import_status = (
                self.ImportStatus.INPUT_ERROR
                if self.result.has_errors()
                or self.result.has_validation_errors()
                else self.ImportStatus.PARSED
            )
            self.parse_finished = timezone.now()
            self.save(
                update_fields=[
                    "import_status",
                    "result",
                    "parse_finished",
                ],
            )
        except Exception as error:
            self.traceback = traceback.format_exc()
            self.error_message = str(error)[:128]
            self.import_status = self.ImportStatus.PARSE_ERROR
            self.save(
                update_fields=[
                    "traceback",
                    "error_message",
                    "import_status",
                ],
            )

    def _parse_data_inner(self) -> Result:
        """Run import process with `dry_run == True`.

        Returns:
            apps.utils.async_import_export.results.Result instance with
            parsing results

        """
        dataset = self._get_data_to_import()
        return self.resource.import_data(
            dataset,
            dry_run=True,
            raise_errors=False,
            collect_failures=True,
        )

    def confirm_import(self):
        """Update task status to IMPORT_CONFIRMED and start parsing.

        This is "intermediate" state between PARSED and IMPORTING and required
        because of possible latency of celery task start.

        Celery task is manually called with `apply_async`, to provide
        possibility of custom `task_id` with which task will be run.

        """
        self._check_import_status_correctness(
            expected_statuses=(self.ImportStatus.PARSED,),
        )

        self.import_status = self.ImportStatus.CONFIRMED
        self.import_task_id = str(uuid.uuid4())
        self.import_started = timezone.now()
        self.save(
            update_fields=[
                "import_status",
                "import_started",
                "import_task_id",
            ],
        )
        transaction.on_commit(self._start_import_data_task)

    def _start_import_data_task(self):
        """Start import task."""
        from .. import tasks

        tasks.import_data_task.apply_async(
            kwargs=dict(job_id=self.pk),
            task_id=self.import_task_id,
        )

    def import_data(self):
        """Import data from `data_file` to DB."""
        expected_status = (
            self.ImportStatus.CREATED
            if self.skip_parse_step
            else self.ImportStatus.CONFIRMED
        )
        self._check_import_status_correctness(
            expected_statuses=(expected_status,),
        )

        self.import_status = self.ImportStatus.IMPORTING
        self.save(update_fields=["import_status"])
        try:
            self.result = self._import_data_inner()
            self.import_status = self.ImportStatus.IMPORTED
            self.import_finished = timezone.now()
            self.save(
                update_fields=[
                    "import_status",
                    "result",
                    "import_finished",
                ],
            )
        except Exception as error:
            self.traceback = traceback.format_exc()
            self.error_message = str(error)[:128]
            self.import_status = self.ImportStatus.IMPORT_ERROR
            self.save(
                update_fields=[
                    "import_status",
                    "error_message",
                    "traceback",
                ],
            )

    def _import_data_inner(self) -> Result:
        """Run import process with saving to DB.

        Transaction is not used as import is slow, so before it finish -
        no instances are saved to DB. So sync works incorrect

        Returns:
            import_export.results.Result instance with parsing results

        """
        data_to_import = self._get_data_to_import()
        return self.resource.import_data(
            data_to_import,
            dry_run=False,
            raise_errors=True,
            use_transactions=True,
            collect_failures=True,
        )

    def _get_import_format_by_ext(
        self,
        file_ext: str,
    ) -> Type[base_formats.Format]:
        """Determine import file format by file extension."""
        supported_formats = self.resource.get_supported_formats()
        for import_format in supported_formats:
            if import_format().get_title().upper() == file_ext.upper().replace(
                ".", "",
            ):
                return import_format
        supported_formats_titles = ",".join(
            supported_format().get_title()
            for supported_format in supported_formats
        )
        raise ValueError(
            f"Incorrect import format: {file_ext}. "
            f"Supported formats: {supported_formats_titles}",
        )

    def _get_data_to_import(self) -> tablib.Dataset:
        """Read ``self.data_file`` content and convert it to dataset."""
        _, file_ext = os.path.splitext(self.data_file.name)
        input_format = self._get_import_format_by_ext(
            file_ext=file_ext,
        )()
        data = self.data_file.read()
        self.data_file.seek(0)
        if not input_format.is_binary():
            data = encoding.force_str(data)
        data_to_import = input_format.create_dataset(data)
        num_rows = len(data_to_import)
        if num_rows > settings.IMPORT_EXPORT_MAX_DATASET_ROWS:
            raise ValueError(
                f"Too many rows `{num_rows}`"
                f"(Max: {settings.IMPORT_EXPORT_MAX_DATASET_ROWS}). "
                f"Input file may be broken. "
                f"If it's spreadsheet file, please delete empty rows.",
            )
        return data_to_import

    def cancel_import(self) -> None:
        """Cancel current data import.

        ImportJob can be CANCELLED only from following states:
            - CREATED
            - PARSING
            - CONFIRMED
            - IMPORTING

        """
        status_task_field_map = {
            self.ImportStatus.CREATED: "parse_task_id",
            self.ImportStatus.PARSING: "parse_task_id",
            self.ImportStatus.CONFIRMED: "import_task_id",
            self.ImportStatus.IMPORTING: "import_task_id",
        }
        self._check_import_status_correctness(
            expected_statuses=status_task_field_map.keys(),  # type: ignore
        )

        # send signal to celery to revoke task
        task_id = getattr(self, status_task_field_map[self.import_status])
        current_app.control.revoke(task_id, terminate=True)

        self.import_status = self.ImportStatus.CANCELLED
        self.save(update_fields=["import_status"])

    def _get_task_state(self, task_id: str) -> TaskStateInfo:
        """Get state info for passed task_id.

        This method may change job status if task failed, but we did not
        save info about this to DB.

        This may happen if task start import/parsing, but processes was killed.
        In that case, job have status `importing`, but it can't be finished.

        """
        async_result = result.AsyncResult(task_id)
        if async_result.state in states.EXCEPTION_STATES:
            # update job's status
            self.import_status = (
                self.ImportStatus.PARSE_ERROR
                if self.import_status == self.ImportStatus.PARSING
                else self.ImportStatus.IMPORT_ERROR
            )
            self.error_message = str(async_result.info)[:128]
            self.traceback = str(async_result.traceback)
            self.save(
                update_fields=[
                    "error_message",
                    "traceback",
                    "import_status",
                ],
            )
            return dict(
                state=async_result.state,
                info={},
            )
        return dict(
            state=async_result.state,
            info=async_result.info,
        )



import os
import re
import typing
import unicodedata
from urllib.parse import unquote_plus

from django.conf import settings
from django.core.files.base import ContentFile
from django.db.models import Q

import requests
from openpyxl.cell.cell import ILLEGAL_CHARACTERS_RE


def normalize_string_value(value: str) -> str:
    """Normalize string value.

    1. Remove leading and trailing whitespaces.
    2. Replace all space characters (' \t\n\r\x0b\x0c') with the Space char.
    3. Remove Unicode C0 controls to prevent problems with the creation of
    XLS(X) files with `openpyxl` lib.
    4. Normalize Unicode string, using `NFKC` form. See the details:
    https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize

    """
    cleaned = " ".join(value.strip().split()).strip()
    return remove_illegal_characters(cleaned)


def get_default_file_mime_type() -> str:
    """Return default MIME type."""
    return settings.MIME_TYPES_MAP[".bin"]


def get_mime_type_by_file_url(file_url: str) -> str:
    """Retrieve file's mime type by file's url according to `MIME_TYPES_MAP`.

    If file extension not in `MIME_TYPES_MAP` it returns default mime type
    for binary files

    """
    file_extension = f".{get_file_extension(file_url)}"
    if file_extension in settings.MIME_TYPES_MAP:
        return settings.MIME_TYPES_MAP[file_extension]

    return get_default_file_mime_type()


def download_file(external_url: str) -> ContentFile:
    """Download file from external resource and return the file object."""
    mime_type = get_mime_type_by_file_url(external_url)
    data = requests.get(external_url)
    file = ContentFile(data.content)
    file.content_type = mime_type
    return file


def get_file_extension(url: str, lower: bool = True) -> str:
    """Extract file extension from path/URL.

    Args:
        url (str): Path to the file
        lower (boolean): Extension in lower
    Returns:
        String: extension of the file (lower or not)

    Example:
        'dir/subdir/file.ext' -> 'ext'

    """
    get_index = url.find("?")
    if get_index != -1:
        url = url[:get_index]  # remove GET params from URL
    ext = os.path.splitext(url)[1][1:]
    return ext.lower() if lower else ext


def remove_illegal_characters(value: str) -> str:
    """Remove `illegal` characters from string values.

    1. Remove Unicode C0 controls to prevent problems with the creation of
    XLS(X) files with `openpyxl` lib.
    2. Normalize Unicode string, using `NFKC` form. See the details:
    https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize

    """
    return unicodedata.normalize(
        "NFKC",
        re.sub(ILLEGAL_CHARACTERS_RE, "", value),
    )


def get_clear_q_filter(str_value: str, attribute_name: str) -> Q:
    """Make clear Q filter for ``str_value``.

    For given string we must create regex pattern that isn't
    dependent on word's cases and extra whitespaces.

    First we build regular expression for ``str_value``
        Example:
            if str_value = 'Hello, world' then regex patter for it
            is r'^Hello\s+world$'

    Then for regex pattern we create Q filter.
        Example:
            If attribute_name is 'title' then Q filter is
            Q(title__iregex=r'^Hello\s+world$')

    Args:
        str_value: some string
        attribute_name: model's attribute name

    """  # noqa: W605
    q_regex_attr = f"{attribute_name}__iregex"

    words = [re.escape(value) for value in str_value.split()]
    pattern = r"\s+".join(words)
    pattern = r"^{0}$".format(pattern)

    return Q(**{q_regex_attr: pattern})


def clean_sequence_of_string_values(
    sequence: typing.Iterable[str],
    ignore_empty: bool = True,
) -> list[str]:
    """Clean sequence of string values.

    Normalize each string value.
    Remove empty items form sequence if `ignore_empty` is True.

    Args:
        sequence: list of strings
        ignore_empty: boolean value which defines should empty strings be
        removed from sequence or not

    Returns:
        cleared_sequence: list of cleared sequence items

    """
    sequence = (normalize_string_value(item) for item in sequence)
    if ignore_empty:
        return list(filter(None, sequence))

    return list(sequence)


def url_to_internal_value(file_url: str) -> str:
    """Convert file url to internal value."""
    file_url = unquote_plus(file_url)
    # manually remove scheme and domain parts, because file name may
    # contain '#' or '?' and they parsed incorrectly
    file_url = file_url.split("://")[-1]
    file_url_components = file_url.split("/")[1:]
    file_url = "/".join(file_url_components)

    if file_url.startswith(settings.MEDIA_URL[1:]):
        # In case of local storing crop the media prefix
        file_url = file_url[len(settings.MEDIA_URL) - 1:]

    elif (
        getattr(settings, "AWS_STORAGE_BUCKET_NAME", None)
        and settings.AWS_STORAGE_BUCKET_NAME in file_url
    ):
        # In case of S3 upload crop S3 bucket name
        file_url = file_url.split(f"{settings.AWS_STORAGE_BUCKET_NAME}/")[-1]

    return file_url



import typing

from django.contrib.contenttypes.models import ContentType
from django.core.handlers.wsgi import WSGIRequest
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from ... import models


class BaseImportExportJobAdminMixin:
    """Mixin provides common methods for ImportJob and ExportJob admins."""

    def has_add_permission(
        self,
        request: WSGIRequest,
        *args,
        **kwargs,
    ) -> bool:
        """Import/Export Jobs should not be created using this interface."""
        return False

    def has_delete_permission(
        self,
        request: WSGIRequest,
        *args,
        **kwargs,
    ) -> bool:
        """Import/Export Jobs should not be deleted using this interface.

        Instead, admins must cancel jobs.

        """
        return False

    def _model(self, obj: models.ImportJob) -> str:
        """Add `model` field of import/export job."""
        try:
            resource_class = import_string(obj.resource_path)
            model = resource_class.Meta.model._meta.verbose_name_plural
        # In case resource has no Meta or model we need to catch AttributeError
        except (ImportError, AttributeError):
            model = _("Unknown")
        return model

    def get_from_content_type(
        self,
        obj: typing.Union[models.ImportJob, models.ExportJob],
    ) -> typing.Union[ContentType, None]:
        """Shortcut to get object from content_type."""
        content_type = obj.resource_kwargs.get("content_type")
        obj_id = obj.resource_kwargs.get("object_id")

        if content_type and obj_id:
            content_type = ContentType.objects.get(id=content_type)
            return content_type.model_class().objects.filter(id=obj_id).first()
        return None



from .serializers.export_job import CreateExportJob, ExportJobSerializer
from .serializers.import_job import CreateImportJob, ImportJobSerializer
from .serializers.progress import ProgressInfoSerializer, ProgressSerializer



import typing

from django.core.exceptions import PermissionDenied
from django.core.handlers.wsgi import WSGIRequest
from django.http import (
    HttpResponse,
    HttpResponseForbidden,
    HttpResponseRedirect,
)
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import re_path, reverse
from django.utils.translation import gettext_lazy as _

from import_export import admin as base_admin
from import_export import forms as base_forms
from import_export import mixins as base_mixins

from ... import models
from . import types


class CeleryExportAdminMixin(
    base_mixins.BaseExportMixin,
    base_admin.ImportExportMixinBase,
):
    """Admin mixin for celery export.

    Admin export work-flow is:
        GET `celery_export_action()` - display form with format type input

        POST `celery_export_action()` - create ExportJob and starts data export
            This view redirects to next view:

        GET `celery_export_job_status_view()` - display ExportJob status (with
            progress bar). When data exporting is done, redirect to next view:

        GET `celery_export_job_results_view()` - display export results. If no
            errors - success message and link to the file with exported data.
            If errors - traceback and error message.

    """
    # export data encoding
    to_encoding = "utf-8"

    # template used to display ExportForm
    celery_export_template_name = "admin/import_export/export.html"

    export_status_template_name = "admin/import_export_extensions/celery_export_status.html"

    export_results_template_name = "admin/import_export_extensions/celery_export_results.html"

    import_export_change_list_template = "admin/import_export/change_list_export.html"

    # Statuses that should be displayed on 'results' page
    export_results_statuses = models.ExportJob.export_finished_statuses

    # Copy methods of mixin from original package to reuse it here
    has_export_permission = base_admin.ExportMixin.has_export_permission

    @property
    def model_info(self) -> types.ModelInfo:
        """Get info of exported model."""
        return types.ModelInfo(
            meta=self.model._meta,
        )

    def get_export_resource_kwargs(self, request, *args, **kwargs):
        """Provide escape settings to resource kwargs."""
        kwargs = super().get_export_resource_kwargs(request, *args, **kwargs)
        kwargs.update({
            "escape_output": self.should_escape_output,
            "escape_html": self.should_escape_html,
            "escape_formulae": self.should_escape_formulae,
        })
        return kwargs

    def get_context_data(
        self,
        request: WSGIRequest,
        **kwargs,
    ) -> dict[str, typing.Any]:
        """Get context data."""
        return {}

    def get_urls(self):
        """Return list of urls.

        /<model/celery-export/:
            ExportForm ('export_action' method)
        /<model>/celery-export/<ID>/:
            status of ExportJob and progress bar ('export_job_status_view')
        /<model>/celery-export/<ID>/results/:
            table with export results (errors)

        """
        urls = super().get_urls()
        export_urls = [
            re_path(
                r"^celery-export/$",
                self.admin_site.admin_view(self.celery_export_action),
                name=f"{self.model_info.app_model_name}_export",
            ),
            re_path(
                r"^celery-export/(?P<job_id>\d+)/$",
                self.admin_site.admin_view(self.export_job_status_view),
                name=(
                    f"{self.model_info.app_model_name}"
                    f"_export_job_status"
                ),
            ),
            re_path(
                r"^celery-export/(?P<job_id>\d+)/results/$",
                self.admin_site.admin_view(
                    self.export_job_results_view,
                ),
                name=(
                    f"{self.model_info.app_model_name}"
                    f"_export_job_results"
                ),
            ),
        ]
        return export_urls + urls

    def celery_export_action(self, request, *args, **kwargs):
        """Show and handle export.

        GET: show export form with format_type input
        POST: create ExportJob instance and redirect to it's status

        """
        if not self.has_export_permission(request):
            raise PermissionDenied

        formats = self.get_export_formats()
        form = base_forms.ExportForm(
            formats,
            request.POST or None,
            resources=self.get_export_resource_classes(),
        )
        resource_kwargs = self.get_export_resource_kwargs(
            request=request,
            *args,
            **kwargs,
        )
        if request.method == "POST" and form.is_valid():
            file_format = formats[int(form.cleaned_data["file_format"])]
            # create ExportJob and redirect to page with it's status
            job = self.create_export_job(
                request=request,
                resource_class=self.choose_export_resource_class(form),
                resource_kwargs=resource_kwargs,
                file_format=file_format,
            )
            return self._redirect_to_export_status_page(
                request=request,
                job=job,
            )

        # GET: display Export Form
        context = self.get_context_data(request)
        context.update(self.admin_site.each_context(request))

        context["title"] = _("Export")
        context["form"] = form
        context["opts"] = self.model_info.meta
        request.current_app = self.admin_site.name

        return TemplateResponse(
            request=request,
            template=[self.celery_export_template_name],
            context=context,
        )

    def export_job_status_view(
        self,
        request: WSGIRequest,
        job_id: int,
        **kwargs,
    ) -> HttpResponse:
        """View to track export job status.

        Displays current export job status and progress (using JS + another
        view).

        If job result is ready - redirects to another page to see results.

        """
        if not self.has_export_permission(request):
            raise PermissionDenied

        job = self.get_export_job(request=request, job_id=job_id)
        if job.export_status in self.export_results_statuses:
            return self._redirect_to_export_results_page(
                request=request,
                job=job,
            )

        context = self.get_context_data(request)
        job_url = reverse("admin:export_job_progress", args=(job.id,))

        context["title"] = _("Export status")
        context["opts"] = self.model_info.meta
        context["export_job"] = job
        context["export_job_url"] = job_url
        request.current_app = self.admin_site.name
        return TemplateResponse(
            request=request,
            template=[self.export_status_template_name],
            context=context,
        )

    def export_job_results_view(
        self,
        request: WSGIRequest,
        job_id: int,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Display export results.

        GET-request:
            * show message
            * if no errors - show file link
            * if errors - show traceback and error

        """
        if not self.has_export_permission(request):
            raise PermissionDenied

        job = self.get_export_job(request=request, job_id=job_id)
        if job.export_status not in self.export_results_statuses:
            return self._redirect_to_export_status_page(
                request=request,
                job=job,
            )

        context = self.get_context_data(request)

        if request.method != "GET":
            return HttpResponseForbidden()

        # GET request, show export results
        context["title"] = _("Export results")
        context["opts"] = self.model._meta
        context["export_job"] = job
        context["result"] = job.export_status

        return TemplateResponse(
            request=request,
            template=[self.export_results_template_name],
            context=context,
        )

    def create_export_job(
        self,
        request: WSGIRequest,
        resource_class: types.ResourceType,
        resource_kwargs: dict[str, typing.Any],
        file_format: types.FormatType,
    ) -> models.ExportJob:
        """Create and return instance of export job with chosen format."""
        job = models.ExportJob.objects.create(
            resource_path=resource_class.class_path,
            resource_kwargs=resource_kwargs,
            file_format_path=".".join(
                [
                    file_format.__module__,
                    file_format.__name__,
                ],
            ),
        )
        return job

    def get_export_job(
        self,
        request: WSGIRequest,
        job_id: int,
    ) -> models.ExportJob:
        """Get ExportJob instance.

        Raises:
            Http404

        """
        return get_object_or_404(models.ExportJob, id=job_id)

    def _redirect_to_export_status_page(
        self,
        request: WSGIRequest,
        job: models.ExportJob,
    ) -> HttpResponse:
        """Shortcut for redirecting to job's status page."""
        url_name = (
            f"admin:{self.model_info.app_model_name}_export_job_status"
        )
        url = reverse(url_name, kwargs=dict(job_id=job.id))
        query = request.GET.urlencode()
        if query:
            url = f"{url}?{query}"
        return HttpResponseRedirect(redirect_to=url)

    def _redirect_to_export_results_page(
        self,
        request: WSGIRequest,
        job: models.ExportJob,
    ) -> HttpResponse:
        """Shortcut for redirecting to job's results page."""
        url_name = (
            f"admin:{self.model_info.app_model_name}_export_job_results"
        )
        url = reverse(url_name, kwargs=dict(job_id=job.id))
        query = request.GET.urlencode()
        if query:
            url = f"{url}?{query}"
        return HttpResponseRedirect(redirect_to=url)

    def changelist_view(
        self,
        request: WSGIRequest,
        context: typing.Optional[dict[str, typing.Any]] = None,
    ):
        """Add the check for permission to changelist template context."""
        context = context or {}
        context["has_export_permission"] = True
        return super().changelist_view(request, context)



import os
import sys

import django

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath("."))
sys.path.append(os.path.abspath(".."))
sys.path.append(os.path.abspath("../tests"))
os.environ["DJANGO_SETTINGS_MODULE"] = "settings"

django.setup()

# -- General configuration ---------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "django-import-export-extensions"
copyright = "2022-2023, Saritasa"
author = "Saritasa"


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
try:
    from import_export_extensions import __version__

    # The short X.Y version.
    version = ".".join(__version__.split(".")[:2])
    # The full version, including alpha/beta/rc tags.
    release = __version__
except ImportError:
    version = release = "dev"


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "django-import-export-extensions"


# -- Options for LaTeX output ------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "django-import-export-extensions.tex",
        "django-import-export-extensions Documentation",
        author,
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "django-import-export-extensions",
        "django-import-export-extensions Documentation",
        [author],
        1,
    ),
]


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "django-import-export-extensions",
        "django-import-export-extensions Documentation",
        author,
        "django-import-export-extensions",
        "Utils for import/export data for Django",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
texinfo_appendices = []

# intersphinx documentation
intersphinx_mapping = {
    "tablib": ("https://tablib.readthedocs.io/en/stable/", None),
    "django-import-export": ("https://django-import-export.readthedocs.io/en/latest/", None),
}



# Generated by Django 4.2.4 on 2023-08-26 04:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('import_export_extensions', '0002_alter_exportjob_export_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='importjob',
            name='skip_parse_step',
            field=models.BooleanField(default=False, help_text='Start importing without confirmation', verbose_name='Skip parse step'),
        ),
    ]



from django import forms
from django.urls import reverse

from ... import models
from ..widgets import ProgressBarWidget


class ExportJobAdminForm(forms.ModelForm):
    """Admin form for `ExportJob` model.

    Adds custom `export_progressbar` field that displays current export
    progress using AJAX requests to specified endpoint. Fields widget is
    defined in `__init__` method.

    """

    export_progressbar = forms.Field(
        label="Export progress",
        required=False,
    )

    def __init__(
        self,
        instance: models.ExportJob,
        *args,
        **kwargs,
    ):
        """Provide `export_progressbar` widget the `ExportJob` instance."""
        super().__init__(*args, instance=instance, **kwargs)
        url_name = "admin:export_job_progress"
        self.fields["export_progressbar"].widget = ProgressBarWidget(
            job=instance,
            url=reverse(url_name, args=(instance.id,)),
        )

    class Meta:
        fields = (
            "export_status",
            "resource_path",
            "file_format_path",
            "data_file",
            "resource_kwargs",
            "traceback",
            "error_message",
            "result",
            "export_task_id",
            "export_started",
            "export_finished",
            "created_by",
            "created",
            "modified",
        )



import typing

from django.contrib import admin, messages
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import QuerySet
from django.http import JsonResponse
from django.urls import re_path
from django.utils.translation import gettext_lazy as _

from ... import models, tasks
from .. import forms
from . import mixins


class ExportJobAdmin(
    mixins.BaseImportExportJobAdminMixin,
    admin.ModelAdmin,
):
    """Admin class for debugging ExportJob."""

    form = forms.ExportJobAdminForm
    exclude = ("result",)
    list_display = (
        "id",
        "export_status",
        "_model",
        "export_started",
        "export_finished",
        "created",
        "created_by",
    )
    list_display_links = (
        "id",
        "export_status",
        "_model",
    )
    export_job_model = models.ExportJob
    list_filter = ("export_status",)
    list_select_related = ("created_by",)
    actions = (
        "cancel_jobs",
    )

    def export_data_action(
        self,
        request: WSGIRequest,
        obj: models.ExportJob,
    ):
        """Admin action for starting data export.

        Data export should auto start after export confirmation. But there
        may be issues with celery and task did not start. So this action
        for such cases.

        """
        tasks.export_data_task.delay(obj.id)

    export_data_action.label = _("Start Export")

    def get_urls(self):
        """Add url to get current export job progress in JSON representation.

        /admin/import_export_extensions/exportjob/<job_id>/progress/

        """
        urls = super().get_urls()
        export_urls = [
            re_path(
                route=r"^(?P<job_id>\d+)/progress/$",
                view=self.admin_site.admin_view(self.export_job_progress_view),
                name="export_job_progress",
            ),
        ]
        return export_urls + urls

    def export_job_progress_view(
        self,
        request: WSGIRequest,
        job_id: int,
        **kwargs,
    ):
        """View to return `ExportJob` status as JSON.

        If current status is exporting, view also returns job state
        and percent of completed work.

        Return:
            Response: dictionary with status (optionally, state and percent).

        """
        try:
            job: models.ExportJob = self.export_job_model.objects.get(
                id=job_id,
            )
        except self.export_job_model.DoesNotExist as error:
            return JsonResponse(dict(validation_error=error.args[0]))

        response_data = dict(status=job.export_status.title())

        if job.export_status != models.ExportJob.ExportStatus.EXPORTING:
            return JsonResponse(response_data)

        percent = 0
        total = 0
        current = 0
        info = job.progress["info"]

        if info and info["total"]:
            percent = int(100 / info["total"] * info["current"])
            total = info["total"]
            current = info["current"]

        response_data.update(
            dict(
                state=job.progress["state"],
                percent=percent,
                total=total,
                current=current,
            ),
        )
        return JsonResponse(response_data)

    def get_readonly_fields(self, request, obj=None):
        """Return readonly fields.

        Some fields are editable for new ExportJob.

        """
        readonly_fields = [
            "export_status",
            "traceback",
            "file_format_path",
            "created",
            "export_started",
            "export_finished",
            "error_message",
            "_model",
        ]
        if obj:
            readonly_fields.extend(
                [
                    "resource_path",
                    "data_file",
                    "resource_kwargs",
                ],
            )

        return readonly_fields

    def get_fieldsets(
        self,
        request: WSGIRequest,
        obj: typing.Optional[models.ExportJob] = None,
    ):
        """Get fieldsets depending on object status."""
        status = (
            _("Status"),
            {
                "fields": (
                    "export_status",
                    "_model",
                    "created",
                    "export_started",
                    "export_finished",
                ),
            },
        )
        progress = (
            _("Status"),
            {
                "fields": (
                    "export_status",
                    "export_progressbar",
                ),
            },
        )
        export_params = (
            _("Export params"),
            {
                "fields": (
                    "resource_path",
                    "resource_kwargs",
                    "file_format_path",
                ),
                "classes": ("collapse",),
            },
        )
        traceback_fields = (
            _("Traceback"),
            {
                "fields": (
                    "error_message",
                    "traceback",
                ),
            },
        )
        result = (
            _("Export results"),
            {
                "fields": ("data_file",),
            },
        )

        if (
            not obj
            or obj.export_status == models.ExportJob.ExportStatus.CREATED
        ):
            return [status, export_params]

        if obj.export_status == models.ExportJob.ExportStatus.EXPORTED:
            return [status, result, export_params]

        if obj.export_status == models.ExportJob.ExportStatus.EXPORTING:
            return [progress, export_params]

        return [status, traceback_fields, export_params]

    @admin.action(description="Cancel selected jobs")
    def cancel_jobs(self, request: WSGIRequest, queryset: QuerySet):
        """Admin action for cancelling data export."""
        for job in queryset:
            try:
                job.cancel_export()
                self.message_user(
                    request,
                    _(f"Export of {job} canceled"),
                    messages.SUCCESS,
                )
            except ValueError as error:
                self.message_user(request, str(error), messages.ERROR)


admin.site.register(models.ExportJob, ExportJobAdmin)



import mimetypes
import typing
from urllib.parse import urlparse

from django.conf import settings
from django.core.exceptions import SuspiciousFileOperation
from django.core.files import File
from django.core.files.storage import default_storage
from django.db.models import Model, Q, QuerySet
from django.forms import ValidationError
from django.utils.encoding import smart_str

from import_export.exceptions import ImportExportError
from import_export.widgets import CharWidget, ManyToManyWidget

from . import utils

DEFAULT_SYSTEM_STORAGE = "django.core.files.storage.FileSystemStorage"


class IntermediateManyToManyWidget(ManyToManyWidget):
    """Widget for M2M field with custom ``through`` model.

    Default M2M widget store just IDs of related objects.
    With intermediate model additional data may be stored.

    This widget (and subclasses) should be used with
    ``IntermediateManyToManyField``. This field expects that widget will return
    plain data for intermediate model as a dict.

    Example models from ``IntermediateManyToManyField`` docstring will be used.

    So this widget's ``clean`` method should return dict with Membership info,
    i.e::

        {
            'band': <Some band>,
            'date_joined': '2015-01-18'
        }

    And ``render`` method should return string that will contain required data.

    """

    def __init__(
        self,
        rem_model: typing.Type[Model],
        instance_separator: str = ",",
        prop_separator: typing.Optional[str] = None,
        rem_field: str = "pk",
        rem_field_lookup: typing.Optional[str] = None,
        extra_fields: typing.Optional[list[str]] = None,
        render_empty: bool = False,
        *args,
        **kwargs,
    ):
        """Init widget.

        Args:
            rem_model(models.Model): remote model (i.e. Band)
            instance_separator(str): separator for instances
            prop_separator(str): separator for instance properties
            rem_field(str): name of remote field to dump remote model (pk
                by default, may be used more human-readable names)
            extra_fields(list[str]): extra fields that should be dumped
                (will be dumped in most dummy way)
            render_empty (bool): defines if render empty values or not
        """
        if prop_separator == ";" or prop_separator is None:
            prop_separator = ":"
        self.rem_model = rem_model
        self.instance_separator = instance_separator
        self.prop_separator = prop_separator
        self.rem_field = rem_field
        self.rem_field_lookup = rem_field_lookup
        self.extra_fields = extra_fields or []
        self.render_empty = render_empty

    def render(self, value: typing.Iterable[Model], *args, **kwargs) -> str:
        """Return an export representation of a intermediate instances.

        For atrists example should be returned something like
            "5:1990-12-12;19:2005-08-16"
            where 5 is band id

        Args:
            value(QuerySet): instances of intermediate model

        """
        instances = []
        for i in value:
            instances.append(
                self.render_instance(i, self._get_related_instance(i)),
            )
        # Clean empty instances
        if not self.render_empty:
            instances = list(filter(None, instances))

        return self.instance_separator.join(instances)

    def render_instance(self, instance: Model, related_object: Model) -> str:
        """Return export representation of one intermediate instance.

        Should take related object PK and extra fields from Intermediate model,
        i.e. Artist.pk and Membership.date_joined

        Args:
            instance: object of intermediate model
            related_object: associated object (i.e. Band)

        """
        # get related object (i.e. Band)
        props = [
            smart_str(self._get_field_value(related_object, self.rem_field)),
        ]

        for attr in self.extra_fields:
            props.append(smart_str(self._get_field_value(instance, attr)))

        return self.prop_separator.join(props)

    def _get_field_value(self, obj: Model, field: str) -> typing.Any:
        """Get `field` value from an object.

        Support chained fields like `field1__field2__field3` which allows to
        get values from obj.field1.field2.field3

        """
        fields_chain = field.split("__")
        value = obj
        for s in fields_chain:
            value = getattr(value, s)

        return value

    def _get_related_instance(self, instance: Model) -> Model:
        """Get related instance based on IntermediateModel instance.

        i.e. get Band based on Membership instance

        Args:
            instance: instance of intermediate model

        """
        for field in instance._meta.get_fields():
            if field.related_model == self.rem_model:
                return getattr(instance, field.name)

    def clean(
        self,
        value: str,
        *args,
        **kwargs,
    ) -> list[dict[str, typing.Any]]:
        """Restore data from dump.

        In ``value`` we have data that saved using ``render`` method. We should
        restore it

        Args:
            value(str): rendered data about instance of intermediate model
            instance:

        Returns:
            list[dict]: parsed data

        Example:
            [{'band': <Band object>, 'date_joined': '1998-12-21'}]

        """
        if not value:
            return []

        # if value is one integer number
        value = str(value)

        # in some cases if click `enter` values `\n\r` inserted
        if self.instance_separator == "\n":
            value = value.replace("\r", "")

        raw_instances = utils.clean_sequence_of_string_values(
            value.split(self.instance_separator),
        )

        result = []
        invalid_instances = []
        restored_objects_ids = []
        for raw_instance in raw_instances:
            try:
                for item in self.clean_instance(raw_instance):
                    if item["object"].pk not in restored_objects_ids:
                        restored_objects_ids.append(item["object"].pk)
                        result.append(item)
            except ValueError:
                invalid_instances.append(raw_instance)

        # if there are entries in `invalid_instances`
        if invalid_instances:
            raise ValueError(
                "You are trying import invalid values: {0}".format(
                    str(invalid_instances),
                ),
            )

        return result

    def clean_instance(self, raw_instance: str) -> list[dict[str, typing.Any]]:
        """Restore info about one instance of intermediate model.

        If there are few instances in DB with same
        `self.rem_field`=`rem_field_value` then it returns few items

        Args:
            raw_instance(str): info about one instance that saved using
                ``render_instance`` method

        Returns:
            list: list of dicts with restored info about one intermediate
                  instance. If there are few instances in DB with same
                  `self.rem_field`=`rem_field_value` then it returns few items

        Example::

            {
                'object': <Band object>,
                'properties': {'date_joined': '2011-10-12'}
            }

        """
        props = raw_instance.split(self.prop_separator)

        if len(props) > len(self.extra_fields) + 1:
            # +1 is for `self.rem_field`
            raise ImportExportError(
                "Too many property separators '{0}' in '{1}'".format(
                    self.prop_separator, raw_instance,
                ),
            )

        # rem instance now contains value used to identify related object,
        # i.e. PK of Band
        # props contain other saved properties of intermediate model
        # i.e. `date_joined`
        rem_field_value, *props = utils.clean_sequence_of_string_values(
            raw_instance.split(self.prop_separator), ignore_empty=False,
        )

        # get related objects
        qs = self.filter_instances(rem_field_value)

        # if we tries import nonexistent instance
        if not qs.exists():
            raise ValueError("Invalid instance {0}".format(raw_instance))

        # build dict with other properties. Ignore extra fields which has
        # empty strings values
        other_props = {
            key: value for key, value in zip(self.extra_fields, props) if value
        }
        return [
            {"object": rem_object, "properties": other_props}
            for rem_object in qs
        ]

    def filter_instances(self, rem_field_value: str) -> QuerySet:
        """Shortcut to filter corresponding instances."""
        if self.rem_field_lookup:
            if self.rem_field_lookup == "regex":
                instance_filter = utils.get_clear_q_filter(
                    rem_field_value, self.rem_field,
                )
            else:
                lookup = f"{self.rem_field}__{self.rem_field_lookup}"
                instance_filter = Q(**{lookup: rem_field_value})
        else:
            instance_filter = Q(**{self.rem_field: rem_field_value})

        return self.rem_model.objects.filter(instance_filter)


class FileWidget(CharWidget):
    """Widget for working with File fields."""

    def __init__(self, filename: str):
        self.filename = filename

    def render(self, value: Model, *args, **kwargs) -> typing.Optional[str]:
        """Convert DB value to URL to file."""
        if not value:
            return None

        if self._get_default_storage() == DEFAULT_SYSTEM_STORAGE:
            return f"http://localhost:8000{value.url}"

        return value.url

    def clean(self, value: str, *args, **kwargs) -> typing.Optional[str]:
        """Get the file and check for exists."""
        if not value:
            return None

        internal_url = utils.url_to_internal_value(urlparse(value).path)

        if not internal_url:
            raise ValidationError("Invalid image path")

        try:
            if default_storage.exists(internal_url):
                return internal_url
        except SuspiciousFileOperation:
            pass

        return self._get_file(value)

    def _get_file(self, url: str) -> File:
        """Download file from the external resource."""
        file = utils.download_file(url)
        ext = mimetypes.guess_extension(file.content_type)
        filename = f"{self.filename}.{ext}" if ext else self.filename

        return File(file, filename)

    def _get_default_storage(self) -> str:
        """Return default system storage used in project.

        Use the value from `STORAGES` if it's available,
        otherwise use `DEFAULT_FILE_STORAGE`.

        `STORAGES` variable replaced `DEFAULT_FILE_STORAGE`, in django 4.2
        https://docs.djangoproject.com/en/4.2/ref/settings/#default-file-storage

        """
        if hasattr(settings, "STORAGES"):
            return settings.STORAGES["default"]["BACKEND"]
        return settings.DEFAULT_FILE_STORAGE



import dataclasses
import typing

from django.db.models.options import Options

from import_export.formats import base_formats

from ... import resources

ResourceObj = typing.Union[
    resources.CeleryResource,
    resources.CeleryModelResource,
]
ResourceType = typing.Type[ResourceObj]
FormatType = typing.Type[base_formats.Format]


@dataclasses.dataclass
class ModelInfo:
    """Contain base info about imported model."""

    meta: Options

    @property
    def name(self) -> str:
        """Get name of model."""
        return self.meta.model_name

    @property
    def app_label(self):
        """App label of model."""
        return self.meta.app_label

    @property
    def app_model_name(self) -> str:
        """Return url name."""
        return f"{self.app_label}_{self.name}"



# Generated by Django 4.1 on 2022-09-01 10:47

import functools

import django.db.models.deletion
import picklefield.fields
from django.conf import settings
from django.db import migrations, models

import import_export_extensions.models.core
import import_export_extensions.models.tools


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="ImportJob",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    import_export_extensions.models.core.CreationDateTimeField(
                        auto_now=True, verbose_name="Created",
                    ),
                ),
                (
                    "modified",
                    import_export_extensions.models.core.ModificationDateTimeField(
                        auto_now=True, verbose_name="Modified",
                    ),
                ),
                (
                    "import_status",
                    models.CharField(
                        choices=[
                            ("CREATED", "Created"),
                            ("PARSING", "Parsing"),
                            ("PARSED", "Parsed"),
                            ("INPUT_ERROR", "Input data error"),
                            ("PARSE_ERROR", "Parse error"),
                            ("CONFIRMED", "Import confirmed"),
                            ("IMPORTING", "Importing"),
                            ("IMPORTED", "Imported"),
                            ("IMPORT_ERROR", "Import error"),
                            ("CANCELLED", "Cancelled"),
                        ],
                        default="CREATED",
                        max_length=20,
                        verbose_name="Job status",
                    ),
                ),
                (
                    "resource_path",
                    models.CharField(
                        help_text="Dotted path to subclass of `import_export.Resource` that should be used for import",
                        max_length=128,
                        verbose_name="Resource class path",
                    ),
                ),
                (
                    "data_file",
                    models.FileField(
                        help_text="File that contain data to be imported",
                        max_length=512,
                        upload_to=functools.partial(
                            import_export_extensions.models.tools.upload_file_to,
                            *(),
                            **{"main_folder_name": "import"}
                        ),
                        verbose_name="Data file",
                    ),
                ),
                (
                    "resource_kwargs",
                    models.JSONField(
                        default=dict,
                        help_text="Keyword parameters required for resource initialization",
                        verbose_name="Resource kwargs",
                    ),
                ),
                (
                    "traceback",
                    models.TextField(
                        blank=True,
                        default=str,
                        help_text="Python traceback in case of parse/import error",
                        verbose_name="Traceback",
                    ),
                ),
                (
                    "error_message",
                    models.CharField(
                        blank=True,
                        default=str,
                        help_text="Python error message in case of parse/import error",
                        max_length=128,
                        verbose_name="Error message",
                    ),
                ),
                (
                    "result",
                    picklefield.fields.PickledObjectField(
                        default=str,
                        editable=False,
                        help_text="Internal import result object that contain info about import statistics. Pickled Python object",
                        verbose_name="Import result",
                    ),
                ),
                (
                    "parse_task_id",
                    models.CharField(
                        default=str,
                        help_text="Celery task ID that start `parse_data`",
                        max_length=36,
                        verbose_name="Parsing task ID",
                    ),
                ),
                (
                    "import_task_id",
                    models.CharField(
                        default=str,
                        help_text="Celery task ID that start `import_data`",
                        max_length=36,
                        verbose_name="Import task ID",
                    ),
                ),
                (
                    "parse_finished",
                    models.DateTimeField(
                        editable=False, null=True, verbose_name="Parse finished",
                    ),
                ),
                (
                    "import_started",
                    models.DateTimeField(
                        editable=False, null=True, verbose_name="Import started",
                    ),
                ),
                (
                    "import_finished",
                    models.DateTimeField(
                        editable=False,
                        null=True,
                        verbose_name="Import finished",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        editable=False,
                        help_text="User which started import",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Created by",
                    ),
                ),
            ],
            options={
                "verbose_name": "Import job",
                "verbose_name_plural": "Import jobs",
            },
        ),
        migrations.CreateModel(
            name="ExportJob",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    import_export_extensions.models.core.CreationDateTimeField(
                        auto_now=True, verbose_name="Created",
                    ),
                ),
                (
                    "modified",
                    import_export_extensions.models.core.ModificationDateTimeField(
                        auto_now=True, verbose_name="Modified",
                    ),
                ),
                (
                    "export_status",
                    models.CharField(
                        choices=[
                            ("CREATED", "Created"),
                            ("EXPORTING", "Exporting"),
                            ("EXPORT_ERROR", "Export Error"),
                            ("EXPORTED", "Exported"),
                        ],
                        default="CREATED",
                        max_length=20,
                        verbose_name="Job status",
                    ),
                ),
                (
                    "resource_path",
                    models.CharField(
                        help_text="Dotted path to subclass of `import_export.Resource` that should be used for export",
                        max_length=128,
                        verbose_name="Resource class path",
                    ),
                ),
                (
                    "file_format_path",
                    models.CharField(
                        help_text="Export file format as path to base_formats class",
                        max_length=128,
                        verbose_name="Export path to class",
                    ),
                ),
                (
                    "data_file",
                    models.FileField(
                        help_text="File that contain exported data",
                        max_length=512,
                        upload_to=functools.partial(
                            import_export_extensions.models.tools.upload_file_to,
                            *(),
                            **{"main_folder_name": "export"}
                        ),
                        verbose_name="Data file",
                    ),
                ),
                (
                    "resource_kwargs",
                    models.JSONField(
                        default=dict,
                        help_text="Keyword parameters required for resource initialization",
                        verbose_name="Resource kwargs",
                    ),
                ),
                (
                    "traceback",
                    models.TextField(
                        blank=True,
                        default=str,
                        help_text="Python traceback in case of export error",
                        verbose_name="Traceback",
                    ),
                ),
                (
                    "error_message",
                    models.CharField(
                        blank=True,
                        default=str,
                        help_text="Python error message in case of export error",
                        max_length=128,
                        verbose_name="Error message",
                    ),
                ),
                (
                    "result",
                    picklefield.fields.PickledObjectField(
                        default=str,
                        editable=False,
                        help_text="Internal export result object that contain info about export statistics. Pickled Python object",
                        verbose_name="Export result",
                    ),
                ),
                (
                    "export_task_id",
                    models.CharField(
                        blank=True,
                        help_text="Celery task ID that start `export_data`",
                        max_length=36,
                        null=True,
                        verbose_name="Export task ID",
                    ),
                ),
                (
                    "export_started",
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        null=True,
                        verbose_name="Export started",
                    ),
                ),
                (
                    "export_finished",
                    models.DateTimeField(
                        blank=True,
                        editable=False,
                        null=True,
                        verbose_name="Export finished",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        editable=False,
                        help_text="User which started export",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Created by",
                    ),
                ),
            ],
            options={
                "verbose_name": "Export job",
                "verbose_name_plural": "Export jobs",
            },
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-25 04:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('import_export_extensions', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='exportjob',
            name='export_status',
            field=models.CharField(choices=[('CREATED', 'Created'), ('EXPORTING', 'Exporting'), ('EXPORT_ERROR', 'Export Error'), ('EXPORTED', 'Exported'), ('CANCELLED', 'Cancelled')], default='CREATED', max_length=20, verbose_name='Job status'),
        ),
    ]



"""This is source code of origin application ``django-import-export``.

The difference is that these forms don't required file format, it is taken from
file. And in ExportForm the default export format is `xlsx`.

"""
import typing

from django import forms
from django.utils.translation import gettext_lazy as _

from import_export.formats import base_formats


class ImportForm(forms.Form):
    """Form for creating import."""

    import_file = forms.FileField(
        label=_("File to import"),
    )


class ExportForm(forms.Form):
    """Form for exporting."""

    file_format = forms.ChoiceField(
        label=_("Choose format"),
        choices=(),
    )

    def __init__(
        self,
        formats: list[typing.Type[base_formats.Format]],
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        choices = []
        initial_choice = 0
        for index, export_format in enumerate(formats):
            extension = export_format().get_title()
            if extension.lower() == "xlsx":
                initial_choice = index
            choices.append((str(index), extension))
        if len(formats) > 1:
            choices.insert(0, ("", "---"))

        self.fields["file_format"].choices = choices
        self.fields["file_format"].initial = initial_choice



import typing

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import PermissionDenied
from django.core.handlers.wsgi import WSGIRequest
from django.forms.forms import Form
from django.http import (
    HttpResponse,
    HttpResponseForbidden,
    HttpResponseRedirect,
)
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import re_path, reverse
from django.utils.translation import gettext_lazy as _

from import_export import admin as base_admin
from import_export import forms as base_forms
from import_export import mixins as base_mixins

from ... import models
from . import types


class CeleryImportAdminMixin(
    base_mixins.BaseImportMixin,
    base_admin.ImportExportMixinBase,
):
    """Admin mixin for celery import.

    Admin import work-flow is:

    GET ``celery_import_action()`` - display form with import file input

    POST ``celery_import_action()`` - save file and create ImportJob.
        This view redirects to next view:

    GET ``celery_import_job_status_view()`` - display ImportJob status (with
        progress bar and critical errors occurred). When data parsing is
        done, redirect to next view:

    GET ``celery_import_job_results_view()`` - display rows that will be
        imported and data parse errors. If no errors - next step.
        If errors - display same form as in ``import_action()``

    POST ``celery_import_job_results_view()`` - start data importing and
        redirect back to GET ``celery_import_job_status_view()``
        with progress bar and import totals.

    """
    # Import data encoding
    from_encoding = "utf-8"

    # Statuses that should be displayed on 'results' page
    results_statuses = models.ImportJob.results_statuses

    # Template used to display ImportForm
    celery_import_template = "admin/import_export/import.html"

    # Template used to display status of import jobs
    import_status_template = "admin/import_export_extensions/celery_import_status.html"

    # template used to display results of import jobs
    import_result_template_name = "admin/import_export_extensions/celery_import_results.html"

    import_export_change_list_template = "admin/import_export/change_list_import.html"

    skip_admin_log = None
    # Copy methods of mixin from original package to reuse it here
    generate_log_entries = base_admin.ImportMixin.generate_log_entries
    get_skip_admin_log = base_admin.ImportMixin.get_skip_admin_log
    has_import_permission = base_admin.ImportMixin.has_import_permission

    @property
    def model_info(self) -> types.ModelInfo:
        """Get info of imported model."""
        return types.ModelInfo(
            meta=self.model._meta,
        )

    def get_context_data(
        self,
        request: WSGIRequest,
        **kwargs,
    ) -> dict[str, typing.Any]:
        """Get context data."""
        return {}

    def get_import_context_data(self, **kwargs):
        """Get context for import data."""
        return self.get_context_data(**kwargs)

    def get_urls(self):
        """Return list of urls.

        * /<model>/<celery-import>/:
            ImportForm ('celery_import_action' method)
        * /<model>/<celery-import>/<ID>/:
            status of ImportJob and progress bar
            ('celery_import_job_status_view')
        * /<model>/<celery-import>/<ID>/results/:
            table with import results (errors) and import confirmation
            ('celery_import_job_results_view')

        """
        urls = super().get_urls()
        import_urls = [
            re_path(
                r"^celery-import/$",
                self.admin_site.admin_view(self.celery_import_action),
                name=f"{self.model_info.app_model_name}_import",
            ),
            re_path(
                r"^celery-import/(?P<job_id>\d+)/$",
                self.admin_site.admin_view(self.celery_import_job_status_view),
                name=(
                    f"{self.model_info.app_model_name}"
                    f"_import_job_status"
                ),
            ),
            re_path(
                r"^celery-import/(?P<job_id>\d+)/results/$",
                self.admin_site.admin_view(
                    self.celery_import_job_results_view,
                ),
                name=(
                    f"{self.model_info.app_model_name}"
                    f"_import_job_results"
                ),
            ),
        ]
        return import_urls + urls

    def celery_import_action(
        self,
        request: WSGIRequest,
        *args,
        **kwargs,
    ):
        """Show and handle ImportForm.

        GET:
            show import form with data_file input form
        POST:
            create ImportJob instance and redirect to it's status

        """
        if not self.has_import_permission(request):
            raise PermissionDenied

        context = self.get_context_data(request)
        resource_classes = self.get_import_resource_classes()

        form = base_forms.ImportForm(
            self.get_import_formats(),
            request.POST or None,
            request.FILES or None,
            resources=resource_classes,
        )
        resource_kwargs = self.get_import_resource_kwargs(request)

        if request.method == "POST" and form.is_valid():
            # create ImportJob and redirect to page with it's status
            resource_class = self.choose_import_resource_class(form)
            job = self.create_import_job(
                request=request,
                resource=resource_class(**resource_kwargs),
                form=form,
            )
            return self._redirect_to_import_status_page(
                request=request,
                job=job,
            )

        # GET: display Import Form
        resources = [
            resource_class(**resource_kwargs)
            for resource_class in resource_classes
        ]

        context.update(self.admin_site.each_context(request))

        context["title"] = _("Import")
        context["form"] = form
        context["opts"] = self.model_info.meta
        context["media"] = self.media + form.media
        context["fields_list"] = [
            (
                resource.get_display_name(),
                [f.column_name for f in resource.get_user_visible_fields()],
            )
            for resource in resources
        ]

        request.current_app = self.admin_site.name
        return TemplateResponse(
            request,
            [self.celery_import_template],
            context,
        )

    def celery_import_job_status_view(
        self,
        request: WSGIRequest,
        job_id: int,
        **kwargs,
    ) -> HttpResponse:
        """View to track import job status.

        Displays current import job status and progress (using JS + another
        view).

        If job result is ready - redirects to another page to see results.

        Also generates admin log entries if the job has `IMPORTED` status.

        """
        if not self.has_import_permission(request):
            raise PermissionDenied

        job = self.get_import_job(request, job_id)
        if job.import_status in self.results_statuses:
            if job.import_status == models.ImportJob.ImportStatus.IMPORTED:
                self.generate_log_entries(job.result, request)
            return self._redirect_to_results_page(
                request=request,
                job=job,
            )

        context = self.get_context_data(request)
        job_url = reverse("admin:import_job_progress", args=(job.id,))
        context.update(
            dict(
                title=_("Import status"),
                opts=self.model_info.meta,
                import_job=job,
                import_job_url=job_url,
            ),
        )
        request.current_app = self.admin_site.name
        return TemplateResponse(
            request=request,
            template=[self.import_status_template],
            context=context,
        )

    def celery_import_job_results_view(
        self,
        request: WSGIRequest,
        job_id: int,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Display table with import results and import confirm form.

        GET-request:
            * show row results
            * if data valid - show import confirmation form
            * if data invalid - show ImportForm for uploading other file

        POST-request:
            * start data importing if data is correct

        """
        if not self.has_import_permission(request):
            raise PermissionDenied

        job = self.get_import_job(request=request, job_id=job_id)
        if job.import_status not in self.results_statuses:
            return self._redirect_to_import_status_page(
                request=request,
                job=job,
            )

        context = self.get_context_data(request=request)

        if request.method == "GET":
            # GET request, show parse results
            result = job.result
            context["import_job"] = job
            context["result"] = result
            context["title"] = _("Import results")

            if job.import_status != models.ImportJob.ImportStatus.PARSED:
                # display import form
                context["import_form"] = base_forms.ImportForm(
                    import_formats=job.resource.SUPPORTED_FORMATS,
                )
            else:
                context["confirm_form"] = Form()

            context.update(self.admin_site.each_context(request))
            context["opts"] = self.model_info.meta
            request.current_app = self.admin_site.name
            return TemplateResponse(
                request,
                [self.import_result_template_name],
                context,
            )

        # POST request. If data is invalid - error
        if job.import_status != models.ImportJob.ImportStatus.PARSED:
            return HttpResponseForbidden(
                "Data invalid, before importing data "
                "needs to be successfully parsed."
                f"Current status: {job.import_status}",
            )

        # start celery task for data importing
        job.confirm_import()
        return self._redirect_to_import_status_page(request=request, job=job)

    def create_import_job(
        self,
        request: WSGIRequest,
        form: Form,
        resource: types.ResourceObj,
    ):
        """Create and return instance of import job."""
        return models.ImportJob.objects.create(
            resource_path=resource.class_path,
            data_file=form.cleaned_data["import_file"],
            resource_kwargs=resource.resource_init_kwargs,
            created_by=request.user,
            skip_parse_step=getattr(settings, "IMPORT_EXPORT_SKIP_ADMIN_CONFIRM", False),
        )

    def get_import_job(
        self,
        request: WSGIRequest,
        job_id: int,
    ) -> models.ImportJob:
        """Get ImportJob instance.

        Raises:
            Http404

        """
        return get_object_or_404(klass=models.ImportJob, id=job_id)

    def _redirect_to_import_status_page(
        self,
        request: WSGIRequest,
        job: models.ImportJob,
    ) -> HttpResponseRedirect:
        """Shortcut for redirecting to job's status page."""
        url_name = (
            f"admin:{self.model_info.app_model_name}_import_job_status"
        )
        url = reverse(url_name, kwargs=dict(job_id=job.id))
        query = request.GET.urlencode()
        url = f"{url}?{query}" if query else url
        return HttpResponseRedirect(redirect_to=url)

    def _redirect_to_results_page(
        self,
        request: WSGIRequest,
        job: models.ImportJob,
    ) -> HttpResponseRedirect:
        """Shortcut for redirecting to job's results page."""
        url_name = (
            f"admin:{self.model_info.app_model_name}_import_job_results"
        )
        url = reverse(url_name, kwargs=dict(job_id=job.id))
        query = request.GET.urlencode()
        url = f"{url}?{query}" if query else url
        if not job.import_status == models.ImportJob.ImportStatus.PARSED:
            return HttpResponseRedirect(redirect_to=url)

        # Redirections add one by one links to `redirect_to`
        key = request.session.get("redirect_key", None)
        session = request.session
        if key:
            links = cache.get(key)
            try:
                session["redirect_to"] = links[0]
                del links[0]
                cache.set(key, links)
            except (TypeError, IndexError):
                session.pop("redirect_to", None)
                session.pop("redirect_key", None)
                cache.delete(key)

        return HttpResponseRedirect(redirect_to=url)

    def changelist_view(
        self,
        request: WSGIRequest,
        context: typing.Optional[dict[str, typing.Any]] = None,
    ):
        """Add the check for permission to changelist template context."""
        context = context or {}
        context["has_import_permission"] = self.has_import_permission(request)
        return super().changelist_view(request, context)



import typing

from rest_framework import request, serializers

from celery import states

from ... import models, resources
from .progress import ProgressSerializer


class ImportProgressSerializer(ProgressSerializer):
    """Serializer to show ImportJob progress."""

    state = serializers.ChoiceField(
        choices=models.ImportJob.ImportStatus.values
        + [
            states.PENDING,
            states.STARTED,
            states.SUCCESS,
        ],
    )


class ImportJobSerializer(serializers.ModelSerializer):
    """Serializer to show information about import job."""

    progress = ImportProgressSerializer()

    class Meta:
        model = models.ImportJob
        fields = (
            "id",
            "import_status",
            "data_file",
            "progress",
            "import_started",
            "import_finished",
            "created",
            "modified",
        )


class CreateImportJob(serializers.Serializer):
    """Base Serializer to start import job.

    It used to set up base workflow of ImportJob creation via API.

    """

    resource_class: typing.Type[resources.CeleryModelResource]

    file = serializers.FileField(required=True)

    def __init__(
        self,
        *args,
        resource_kwargs: typing.Optional[dict[str, typing.Any]] = None,
        **kwargs,
    ):
        """Set filter kwargs and current user."""
        super().__init__(*args, **kwargs)
        self._request: request.Request = self.context.get("request")
        self._resource_kwargs = resource_kwargs or {}
        self._user = getattr(self._request, "user", None)

    def create(
        self,
        validated_data: dict[str, typing.Any],
    ) -> models.ImportJob:
        """Create import job."""
        return models.ImportJob.objects.create(
            data_file=validated_data["file"],
            resource_path=self.resource_class.class_path,
            resource_kwargs=self._resource_kwargs,
            created_by=self._user,
        )

    def update(self, instance, validated_data):
        """Empty method to pass linters checks."""


def get_create_import_job_serializer(
    resource: typing.Type[resources.CeleryModelResource],
) -> typing.Type:  # type: ignore
    """Create serializer for ImportJobs creation."""

    class _CreateImportJob(CreateImportJob):
        """Serializer to start import job."""

        resource_class: typing.Type[resources.CeleryModelResource] = resource

    return type(
        f"{resource.__name__}CreateImportJob",
        (_CreateImportJob,),
        {},
    )



import typing

from rest_framework import (
    decorators,
    mixins,
    permissions,
    response,
    status,
    viewsets,
)
from rest_framework.exceptions import ValidationError
from rest_framework.request import Request

import django_filters

from ... import models, resources
from .. import serializers


class ExportBase(type):
    """Add custom create action for each ExportJobViewSet."""

    def __new__(cls, name, bases, attrs, **kwargs):
        """Dynamically create an export start api endpoint.

        We need this to specify on fly action's filterset_class and queryset
        (django-filters requires view's queryset and filterset_class's
        queryset model to match). Also, if drf-spectacular is installed
        specify request and response, and enable filters.

        """
        viewset: typing.Type["ExportJobViewSet"] = super().__new__(
            cls,
            name,
            bases,
            attrs,
            **kwargs,
        )
        # Skip if it is a base viewset, since none of needed class attrs are
        # specified
        if name == "ExportJobViewSet":
            return viewset

        def start(self: "ExportJobViewSet", request: Request):
            """Validate request data and start ExportJob."""
            serializer = self.get_serializer(
                data=request.data,
                filter_kwargs=request.query_params,
            )
            serializer.is_valid(raise_exception=True)
            export_job = serializer.save()
            return response.Response(
                data=self.get_detail_serializer_class()(
                    instance=export_job,
                ).data,
                status=status.HTTP_201_CREATED,
            )

        viewset.start = decorators.action(
            methods=["POST"],
            detail=False,
            queryset=viewset.resource_class.get_model_queryset(),
            filterset_class=getattr(
                viewset.resource_class, "filterset_class", None,
            ),
            filter_backends=[
                django_filters.rest_framework.DjangoFilterBackend,
            ],
        )(start)
        # Correct specs of drf-spectacular if it is installed
        try:
            from drf_spectacular.utils import extend_schema, extend_schema_view

            detail_serializer_class = viewset().get_detail_serializer_class()
            return extend_schema_view(
                start=extend_schema(
                    filters=True,
                    request=viewset().get_export_create_serializer_class(),
                    responses={
                        status.HTTP_201_CREATED: detail_serializer_class,
                    },
                ),
                cancel=extend_schema(
                    request=None,
                    responses={
                        status.HTTP_200_OK: detail_serializer_class,
                    },
                ),
            )(viewset)
        except ImportError:
            pass
        return viewset


class ExportJobViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
    metaclass=ExportBase,
):
    """Base API viewset for ExportJob model.

    Based on resource_class it will generate an endpoint which will allow to
    start an export of model which was specified in resource_class. This
    endpoint will support filtration based on FilterSet class specified in
    resource. On success this endpoint we return an instance of export, to
    get status of job, just use detail(retrieve) endpoint.

    """

    permission_classes = (permissions.IsAuthenticated,)
    queryset = models.ExportJob.objects.all()
    serializer_class = serializers.ExportJobSerializer
    resource_class: typing.Optional[
        typing.Type[resources.CeleryModelResource]
    ] = None
    filterset_class: django_filters.rest_framework.FilterSet = None

    def get_queryset(self):
        """Filter export jobs by resource used in viewset."""
        return super().get_queryset().filter(
            resource_path=self.resource_class.class_path,
        )

    def get_resource_kwargs(self) -> dict[str, typing.Any]:
        """Provide extra arguments to resource class."""
        return {}

    def get_serializer(self, *args, **kwargs):
        """Provide resource kwargs to serializer class."""
        if self.action == "start":
            kwargs.setdefault("resource_kwargs", self.get_resource_kwargs())
        return super().get_serializer(*args, **kwargs)

    def get_serializer_class(self):
        """Return special serializer on creation."""
        if self.action == "start":
            return self.get_export_create_serializer_class()
        return self.get_detail_serializer_class()

    def get_detail_serializer_class(self):
        """Get serializer which will be used show details of export job."""
        return self.serializer_class

    def get_export_create_serializer_class(self):
        """Get serializer which will be used to start export job."""
        return serializers.get_create_export_job_serializer(
            self.resource_class,
        )

    @decorators.action(methods=["POST"], detail=True)
    def cancel(self, *args, **kwargs):
        """Cancel export job that is in progress."""
        job: models.ExportJob = self.get_object()

        try:
            job.cancel_export()
        except ValueError as error:
            raise ValidationError(error.args[0])

        serializer = self.get_serializer(instance=job)
        return response.Response(
            status=status.HTTP_200_OK,
            data=serializer.data,
        )



import functools
import uuid


def upload_file_to(
    instance,
    filename: str,
    main_folder_name: str,
) -> str:
    """Upload instance's `file` to unique folder.

    Args:
        instance (typing.Union[ImportJob, ExportJob]): instance of job model
        main_folder_name(str): main folder -> import or export
        filename (str): file name of document's file

    Returns:
        str: Generated path for document's file.

    """
    return (
        "import_export_extensions/"
        f"{main_folder_name}/{uuid.uuid4()}/{filename}"
    )


upload_export_file_to = functools.partial(
    upload_file_to,
    main_folder_name="export",
)
upload_import_file_to = functools.partial(
    upload_file_to,
    main_folder_name="import",
)



from rest_framework import serializers


class ProgressInfoSerializer(serializers.Serializer):
    """Serializer to show progress info, like how much is done."""

    current = serializers.IntegerField()
    total = serializers.IntegerField()


class ProgressSerializer(serializers.Serializer):
    """Serializer to show progress of job."""

    info = ProgressInfoSerializer()




from django import forms


class ProgressBarWidget(forms.Widget):
    """Widget for progress bar field.

    Value for `progress_bar` element is changed using JS code.

    """

    def __init__(self, *args, **kwargs):
        """Get ``ImportJob`` or ``ExportJob`` instance from kwargs.

        ``ImportJob`` or ``ExportJob`` instance is used
        to render hidden element in `render` method.

        """
        self.job = kwargs.pop("job")
        self.url = kwargs.pop("url")
        super().__init__(*args, **kwargs)

    def render(self, *args, **kwargs) -> str:
        """Render HTML5 `progress` element.

        Additionally, method provides hidden `import_job_url` and
        `export_job_url` value that is used in `js/admin/progress_bar.js`
        to send GET requests.

        """
        progress_bar = f"""
            <progress
                value="0"
                max="100"
                id="progress-bar"
                data-url="{self.url}">
            </progress>
        """

        return progress_bar

    class Media:
        """Class with custom assets for widget."""

        css = dict(
            all=("import_export_extensions/css/widgets/progress_bar.css",),
        )
        js = (
            "admin/js/jquery.init.js",
            "import_export_extensions/js/widgets/progress_bar.js",
        )



import typing

from rest_framework import (
    decorators,
    mixins,
    permissions,
    response,
    status,
    viewsets,
)
from rest_framework.exceptions import ValidationError

from ... import models, resources
from .. import serializers


class ImportBase(type):
    """Add custom create action for each ImportJobViewSet."""

    def __new__(cls, name, bases, attrs, **kwargs):
        """Dynamically create an import start api endpoint.

        If drf-spectacular is installed
        specify request and response, and enable filters.

        """
        viewset: typing.Type["ImportJobViewSet"] = super().__new__(
            cls,
            name,
            bases,
            attrs,
            **kwargs,
        )
        # Skip if it is a base viewset, since none of needed class attrs are
        # specified
        if name == "ImportJobViewSet":
            return viewset

        # Correct specs of drf-spectacular if it is installed
        try:
            from drf_spectacular.utils import extend_schema, extend_schema_view

            detail_serializer_class = viewset().get_detail_serializer_class()
            return extend_schema_view(
                start=extend_schema(
                    request=viewset().get_import_create_serializer_class(),
                    responses={
                        status.HTTP_201_CREATED: detail_serializer_class,
                    },
                ),
                confirm=extend_schema(
                    request=None,
                    responses={
                        status.HTTP_200_OK: detail_serializer_class,
                    },
                ),
                cancel=extend_schema(
                    request=None,
                    responses={
                        status.HTTP_200_OK: detail_serializer_class,
                    },
                ),
            )(viewset)
        except ImportError:
            pass
        return viewset


class ImportJobViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    viewsets.GenericViewSet,
    metaclass=ImportBase,
):
    """Base API viewset for ImportJob model.

    Based on resource_class it will generate an endpoint which will allow to
    start an import to model which was specified in resource_class. On success
    this endpoint we return an instance of import.

    Endpoints:
        list - to get list of all import jobs
        details(retrieve) - to get status of import job
        start - create import job and start parsing data from attached file
        confirm - confirm import after parsing process is finished
        cancel - stop importing/parsing process and cancel this import job

    """

    permission_classes = (permissions.IsAuthenticated,)
    queryset = models.ImportJob.objects.all()
    serializer_class = serializers.ImportJobSerializer
    resource_class: typing.Optional[
        typing.Type[resources.CeleryModelResource]
    ] = None

    def get_queryset(self):
        """Filter import jobs by resource used in viewset."""
        return super().get_queryset().filter(
            resource_path=self.resource_class.class_path,
        )

    def get_resource_kwargs(self) -> dict[str, typing.Any]:
        """Provide extra arguments to resource class."""
        return {}

    def get_serializer(self, *args, **kwargs):
        """Provide resource kwargs to serializer class."""
        if self.action == "start":
            kwargs.setdefault("resource_kwargs", self.get_resource_kwargs())
        return super().get_serializer(*args, **kwargs)

    def get_serializer_class(self):
        """Return special serializer on creation."""
        if self.action == "start":
            return self.get_import_create_serializer_class()
        return self.get_detail_serializer_class()

    def get_detail_serializer_class(self):
        """Get serializer which will be used show details of import job."""
        return self.serializer_class

    def get_import_create_serializer_class(self):
        """Get serializer which will be used to start import job."""
        return serializers.get_create_import_job_serializer(
            self.resource_class,
        )

    @decorators.action(methods=["POST"], detail=False)
    def start(self, request, *args, **kwargs):
        """Validate request data and start ImportJob."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        import_job = serializer.save()

        return response.Response(
            data=self.get_detail_serializer_class()(
                instance=import_job,
            ).data,
            status=status.HTTP_201_CREATED,
        )

    @decorators.action(methods=["POST"], detail=True)
    def confirm(self, *args, **kwargs):
        """Confirm import job that has `parsed` status."""
        job: models.ImportJob = self.get_object()

        try:
            job.confirm_import()
        except ValueError:
            raise ValidationError(
                f"Wrong import job status: {job.import_status}",
            )

        serializer = self.get_serializer(instance=job)
        return response.Response(
            status=status.HTTP_200_OK,
            data=serializer.data,
        )

    @decorators.action(methods=["POST"], detail=True)
    def cancel(self, *args, **kwargs):
        """Cancel import job that is in progress."""
        job: models.ImportJob = self.get_object()

        try:
            job.cancel_import()
        except ValueError:
            raise ValidationError(
                f"Wrong import job status: {job.import_status}",
            )

        serializer = self.get_serializer(instance=job)
        return response.Response(
            status=status.HTTP_200_OK,
            data=serializer.data,
        )



import typing

from rest_framework import request, serializers

from celery import states
from django_filters.utils import translate_validation

from ... import models, resources
from .progress import ProgressSerializer


class ExportProgressSerializer(ProgressSerializer):
    """Serializer to show ExportJob progress."""

    state = serializers.ChoiceField(
        choices=models.ExportJob.ExportStatus.values
        + [
            states.PENDING,
            states.STARTED,
            states.SUCCESS,
        ],
    )


class ExportJobSerializer(serializers.ModelSerializer):
    """Serializer to show information about export job."""

    progress = ExportProgressSerializer()

    class Meta:
        model = models.ExportJob
        fields = (
            "id",
            "export_status",
            "data_file",
            "progress",
            "export_started",
            "export_finished",
            "created",
            "modified",
        )


class CreateExportJob(serializers.Serializer):
    """Base Serializer to start export job.

    It used to set up base workflow of ExportJob creation via API.

    """

    resource_class: typing.Type[resources.CeleryModelResource]

    def __init__(
        self,
        *args,
        filter_kwargs: typing.Optional[dict[str, typing.Any]] = None,
        resource_kwargs: typing.Optional[dict[str, typing.Any]] = None,
        **kwargs,
    ):
        """Set filter kwargs and current user."""
        super().__init__(*args, **kwargs)
        self._filter_kwargs = filter_kwargs
        self._resource_kwargs = resource_kwargs or {}
        self._request: request.Request = self.context.get("request")
        self._user = getattr(self._request, "user", None)

    def validate(self, attrs: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Check that filter kwargs are valid."""
        if not self._filter_kwargs:
            return attrs

        filter_instance = self.resource_class.filterset_class(
            data=self._filter_kwargs,
        )
        if not filter_instance.is_valid():
            raise translate_validation(error_dict=filter_instance.errors)
        return attrs

    def create(
        self,
        validated_data: dict[str, typing.Any],
    ) -> models.ExportJob:
        """Create export job."""
        file_format_class = self.resource_class.get_supported_extensions_map()[
            validated_data["file_format"]
        ]
        return models.ExportJob.objects.create(
            resource_path=self.resource_class.class_path,
            file_format_path=f"{file_format_class.__module__}.{file_format_class.__name__}",
            resource_kwargs=dict(
                filter_kwargs=self._filter_kwargs,
                **self._resource_kwargs,
            ),
            created_by=self._user,
        )

    def update(self, instance, validated_data):
        """Empty method to pass linters checks."""


def get_create_export_job_serializer(
    resource: typing.Type[resources.CeleryModelResource],
) -> typing.Type:  # type: ignore
    """Create serializer for ExportJobs creation."""

    class _CreateExportJob(CreateExportJob):
        """Serializer to start export job."""

        resource_class: typing.Type[resources.CeleryModelResource] = resource
        file_format = serializers.ChoiceField(
            required=True,
            choices=[
                supported_format().get_extension()
                for supported_format in resource.SUPPORTED_FORMATS
            ],
        )

    return type(
        f"{resource.__name__}CreateExportJob",
        (_CreateExportJob,),
        {},
    )



import mimetypes

from django.apps import AppConfig
from django.conf import settings
from django.utils.translation import gettext_lazy as _

# Default configuration
# Maximum num of rows to be imported
DEFAULT_MAX_DATASET_ROWS = 100000


class CeleryImportExport(AppConfig):
    """Default configuration for CeleryImportExport app."""

    name = "import_export_extensions"
    verbose_name = _("Celery Import/Export")

    def ready(self):
        """Set up default settings."""
        settings.IMPORT_EXPORT_MAX_DATASET_ROWS = getattr(
            settings,
            "IMPORT_EXPORT_MAX_DATASET_ROWS",
            DEFAULT_MAX_DATASET_ROWS,
        )
        settings.MIME_TYPES_MAP = mimetypes.types_map.copy()



import typing

from django.contrib import admin, messages
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import QuerySet
from django.http import JsonResponse
from django.template.loader import render_to_string
from django.urls import re_path
from django.utils.translation import gettext_lazy as _

from ... import models
from .. import forms
from . import mixins


class ImportJobAdmin(
    mixins.BaseImportExportJobAdminMixin,
    admin.ModelAdmin,
):
    """Admin class for debugging ImportJob."""

    form = forms.ImportJobAdminForm
    exclude = ("result",)
    list_display = (
        "id",
        "import_status",
        "_model",
        "created_by",
        "created",
        "parse_finished",
        "import_started",
        "import_finished",
    )
    list_display_links = (
        "id",
        "import_status",
        "_model",
    )
    import_job_model = models.ImportJob
    list_filter = ("import_status",)
    list_select_related = ("created_by",)
    actions = (
        "cancel_jobs",
        "confirm_jobs",
    )

    def get_queryset(self, request: WSGIRequest):
        """Override `get_queryset`.

        Do not get `result` from db because it can be rather big and is not
        used in admin.

        """
        return super().get_queryset(request).defer("result")

    def get_urls(self):
        """Add url to get current import job progress in JSON representation.

        /admin/import_export_extensions/importjob/<job_id>/progress/

        """
        urls = super().get_urls()
        import_urls = [
            re_path(
                r"^(?P<job_id>\d+)/progress/$",
                self.admin_site.admin_view(self.import_job_progress_view),
                name="import_job_progress",
            ),
        ]
        return import_urls + urls

    def import_job_progress_view(
        self,
        request: WSGIRequest,
        job_id: int,
        **kwargs,
    ) -> JsonResponse:
        """View to return ``ImportJob`` status as JSON.

        If current status is parsing/importing, view also returns job state
        and percent of completed work.

        Return:
            Response: dictionary with status (optionally, state and percent).

        """
        try:
            job: models.ImportJob = self.import_job_model.objects.get(
                id=job_id,
            )
        except self.import_job_model.DoesNotExist as error:
            return JsonResponse(dict(validation_error=error.args[0]))

        response_data = dict(status=job.import_status.title())

        if job.import_status in models.ImportJob.progress_statuses:
            percent = 0
            total = 0
            current = 0
            info = job.progress["info"]

            if info and info["total"]:
                percent = int(100 / info["total"] * info["current"])
                total = info["total"]
                current = info["current"]

            response_data.update(
                dict(
                    state=job.progress["state"],
                    percent=percent,
                    total=total,
                    current=current,
                ),
            )

        return JsonResponse(response_data)

    def get_readonly_fields(
        self,
        request: WSGIRequest,
        obj: typing.Optional[models.ImportJob] = None,
    ) -> list[str]:
        """Return readonly fields.

        Some fields are editable for new ImportJob.

        """
        readonly_fields = [
            "import_status",
            "_model",
            "created_by",
            "traceback_str",
            "_show_results",
            "_input_errors",
            "created",
            "parse_finished",
            "import_started",
            "import_finished",
        ]
        if obj:
            readonly_fields.extend(
                [
                    "resource_path",
                    "data_file",
                    "resource_kwargs",
                ],
            )

        return readonly_fields

    def _show_results(
        self,
        obj: typing.Optional[models.ImportJob] = None,
    ) -> str:
        """Show results totals.

        Example return value:
        New: 99
        Update: 1
        Delete: 0
        Skip: 0
        Error: 0

        """
        if not obj:
            return ""

        result_sections = []
        for key, value in obj.result.totals.items():
            status_template = f"{key.title()}: {value}"
            result_sections.append(status_template)

        return "\n".join(result_sections)

    _show_results.short_description = _("Parse/Import results")

    def _input_errors(self, job: models.ImportJob):
        """Render html with input errors."""
        template = "admin/import_export_extensions/import_job_results.html"
        return render_to_string(
            template,
            dict(result=job.result),
        )

    _input_errors.short_description = "Import data"
    _input_errors.allow_tags = True

    def get_fieldsets(
        self,
        request: WSGIRequest,
        obj: typing.Optional[models.ImportJob] = None,
    ):
        """Get fieldsets depending on object status."""
        status = (
            _("Status"),
            {
                "fields": (
                    "import_status",
                    "_model",
                    "created_by",
                    "created",
                    "parse_finished",
                    "import_started",
                    "import_finished",
                ),
            },
        )
        progress = (
            _("Status"),
            {
                "fields": (
                    "import_status",
                    "import_progressbar",
                ),
            },
        )
        import_params = (
            _("Import params"),
            {
                "fields": (
                    "data_file",
                    "resource_path",
                    "resource_kwargs",
                ),
                "classes": ("collapse",),
            },
        )
        traceback_ = (
            _("Traceback"),
            {
                "fields": ("traceback",),
            },
        )
        result = (
            _("Result totals"),
            {
                "fields": ("_show_results",),
                "classes": ("collapse",),
            },
        )
        data = (
            _("Importing data"),
            {
                "fields": ("_input_errors",),
                "classes": ("collapse",),
            },
        )

        if not obj:
            return [status, import_params]

        if obj.import_status == models.ImportJob.ImportStatus.CREATED:
            return [status, import_params]

        if obj.import_status in models.ImportJob.results_statuses:
            return [status, result, data, import_params]

        if obj.import_status in models.ImportJob.progress_statuses:
            return [status, progress, import_params]

        return [status, traceback_, import_params]

    @admin.action(description="Cancel selected jobs")
    def cancel_jobs(self, request: WSGIRequest, queryset: QuerySet):
        """Admin action for cancelling data import."""
        for job in queryset:
            try:
                job.cancel_import()
                self.message_user(
                    request,
                    _(f"Import of {job} canceled"),
                    messages.SUCCESS,
                )
            except ValueError as error:
                self.message_user(request, str(error), messages.ERROR)

    @admin.action(description="Confirm selected jobs")
    def confirm_jobs(self, request: WSGIRequest, queryset: QuerySet):
        """Admin action for confirming data import."""
        for job in queryset:
            try:
                job.confirm_import()
                self.message_user(
                    request,
                    _(f"Import of {job} confirmed"),
                    messages.SUCCESS,
                )
            except ValueError as error:
                self.message_user(request, str(error), messages.ERROR)


admin.site.register(models.ImportJob, ImportJobAdmin)



from celery import shared_task

from . import models


@shared_task()
def parse_data_task(job_id: int):
    """Async task for starting data parsing."""
    models.ImportJob.objects.get(pk=job_id).parse_data()


@shared_task()
def import_data_task(job_id: int):
    """Async task for starting data import."""
    models.ImportJob.objects.get(pk=job_id).import_data()


@shared_task()
def export_data_task(job_id: int):
    """Async task for starting data export."""
    job = models.ExportJob.objects.get(id=job_id)
    job.export_data()



import typing
from enum import Enum

from django.db.models import QuerySet
from django.utils import timezone
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

import tablib
from celery import current_task, result
from django_filters import rest_framework as filters
from django_filters.utils import translate_validation
from import_export import resources
from import_export.formats import base_formats
from import_export.results import Error as BaseError


class Error(BaseError):
    """Customization of over base Error class from import export."""

    def __repr__(self) -> str:
        """Return object representation in string format."""
        return f"Error({self.error})"

    def __reduce__(self):
        """Simplify Exception object for pickling.

        `error` object may contain not pickable objects (for example, django's
        lazy text), so here it replaced with simple string.

        """
        self.error = str(self.error)
        return super().__reduce__()


class TaskState(Enum):
    """Class with possible task state values."""
    IMPORTING = _("Importing")
    EXPORTING = _("Exporting")
    PARSING = _("Parsing")


class CeleryResourceMixin:
    """Mixin for resources for background import/export using celery."""
    filterset_class: typing.Type[filters.FilterSet]
    SUPPORTED_FORMATS: list[
        typing.Type[base_formats.Format]
    ] = base_formats.DEFAULT_FORMATS

    def __init__(
        self,
        filter_kwargs: typing.Optional[dict[str, typing.Any]] = None,
        **kwargs,
    ):
        """Remember init kwargs."""
        self._filter_kwargs = filter_kwargs
        self.resource_init_kwargs: dict[str, typing.Any] = kwargs
        super().__init__()

    def get_queryset(self):
        """Filter export queryset via filterset class."""
        queryset = super().get_queryset()
        if not self._filter_kwargs:
            return queryset
        filter_instance = self.filterset_class(
            data=self._filter_kwargs,
        )
        if not filter_instance.is_valid():
            raise translate_validation(filter_instance.errors)
        return filter_instance.filter_queryset(queryset=queryset)

    @classproperty
    def class_path(cls) -> str:
        """Get path of class to import it."""
        return ".".join([cls.__module__, cls.__name__])

    @classmethod
    def get_supported_formats(cls) -> list[typing.Type[base_formats.Format]]:
        """Get a list of supported formats."""
        return cls.SUPPORTED_FORMATS

    @classmethod
    def get_supported_extensions_map(cls) -> dict[
        str, typing.Type[base_formats.Format],
    ]:
        """Get a map of supported formats and their extensions."""
        return {
            supported_format().get_extension(): supported_format
            for supported_format in cls.SUPPORTED_FORMATS
        }

    def import_data(
        self,
        dataset: tablib.Dataset,
        dry_run: bool = False,
        raise_errors: bool = False,
        use_transactions: typing.Optional[bool] = None,
        collect_failed_rows: bool = False,
        rollback_on_validation_errors: bool = False,
        **kwargs,
    ):
        """Init task state before importing."""
        self.initialize_task_state(
            state=(
                TaskState.IMPORTING.name if not dry_run
                else TaskState.PARSING.name
            ),
            queryset=dataset,
        )
        return super().import_data(  # type: ignore
            dataset=dataset,
            dry_run=dry_run,
            raise_errors=raise_errors,
            use_transactions=use_transactions,
            collect_failed_rows=collect_failed_rows,
            rollback_on_validation_errors=rollback_on_validation_errors,
            **kwargs,
        )

    def import_row(
        self,
        row,
        instance_loader,
        using_transactions=True,
        dry_run=False,
        raise_errors=False,
        **kwargs,
    ):
        """Update task status as we import rows."""
        imported_row = super().import_row(
            row=row,
            instance_loader=instance_loader,
            using_transactions=using_transactions,
            dry_run=dry_run,
            raise_errors=raise_errors,
            **kwargs,
        )
        self.update_task_state(
            state=(
                TaskState.IMPORTING.name if not dry_run
                else TaskState.PARSING.name
            ),
        )
        return imported_row

    def export(
        self,
        queryset: QuerySet = None,
        *args,
        **kwargs,
    ) -> tablib.Dataset:
        """Init task state before exporting."""
        if queryset is None:
            queryset = self.get_queryset()
        self.initialize_task_state(
            state=TaskState.EXPORTING.name,
            queryset=queryset,
        )
        return super().export(  # type: ignore
            queryset=queryset,
            *args,
            **kwargs,
        )

    def export_resource(self, obj):
        """Update task status as we export rows."""
        resource = [
            self.export_field(field, obj) for field in self.get_export_fields()
        ]
        self.update_task_state(state=TaskState.EXPORTING.name)
        return resource

    def initialize_task_state(
        self,
        state: str,
        queryset: typing.Union[QuerySet, tablib.Dataset],
    ):
        """Set initial state of the task to track progress.

        Counts total number of instances to import/export and
        generate state for the task.

        """
        if not current_task or current_task.request.called_directly:
            return

        if isinstance(queryset, QuerySet):
            total = queryset.count()
        else:
            total = len(queryset)

        self._update_current_task_state(
            state=state,
            meta=dict(
                current=0,
                total=total,
            ),
        )

    def update_task_state(
        self,
        state: str,
    ):
        """Update state of the current event.

        Receives meta of the current task and increase the `current`
        field by 1.

        """
        if not current_task or current_task.request.called_directly:
            return

        async_result = result.AsyncResult(current_task.request.get("id"))
        if not async_result.result:
            return

        self._update_current_task_state(
            state=state,
            meta=dict(
                current=async_result.result.get("current", 0) + 1,
                total=async_result.result.get("total", 0),
            ),
        )

    def _update_current_task_state(self, state: str, meta: dict[str, int]):
        """Update state of task where resource is executed."""
        current_task.update_state(
            state=state,
            meta=meta,
        )

    def generate_export_filename(self, file_format: base_formats.Format):
        """Generate export filename."""
        return self._generate_export_filename_from_model(file_format)

    def _generate_export_filename_from_model(
        self,
        file_format: base_formats.Format,
    ):
        """Generate export file name from model name."""
        model = self._meta.model._meta.verbose_name_plural
        date_str = timezone.now().strftime("%Y-%m-%d")
        extension = file_format.get_extension()
        return f"{model}-{date_str}.{extension}"

    @classmethod
    def get_error_result_class(cls):
        """Override default error class."""
        return Error


class CeleryResource(CeleryResourceMixin, resources.Resource):
    """Resource which supports importing via celery."""


class CeleryModelResource(CeleryResourceMixin, resources.ModelResource):
    """ModelResource which supports importing via celery."""

    @classmethod
    def get_model_queryset(cls):
        """Return a queryset of all objects for this model.

        Override this if you
        want to limit the returned queryset.

        Same as resources.ModelResource get_queryset.

        """
        return cls._meta.model.objects.all()



import typing

from django.db import models
from django.utils.translation import gettext_lazy as _


class CreationDateTimeField(models.DateTimeField):
    """DateTimeField to indicate created datetime.

    By default, sets editable=False, blank=True, auto_now_add=True.

    """

    def __init__(self, **kwargs):
        super().__init__(
            auto_now=True,
            verbose_name=_("Created"),
        )


class ModificationDateTimeField(models.DateTimeField):
    """DateTimeField  to indicate modified datetime.

    By default, sets editable=False, blank=True, auto_now=True.

    Sets value to now every time the object is saved.

    """

    def __init__(self, **kwargs):
        super().__init__(
            auto_now=True,
            verbose_name=_("Modified"),
        )


class TimeStampedModel(models.Model):
    """Model which tracks creation and modification time.

    An abstract base class model that provides self-managed "created" and
    "modified" fields.

    """

    created = CreationDateTimeField()
    modified = ModificationDateTimeField()

    class Meta:
        abstract = True


class TaskStateInfo(typing.TypedDict):
    """Class representing task state dict."""
    state: str
    info: typing.Optional[dict[str, int]]



import pathlib
import traceback
import typing
import uuid

from django.conf import settings
from django.core import files as django_files
from django.db import models, transaction
from django.utils import encoding, module_loading, timezone
from django.utils.translation import gettext_lazy as _

from celery import current_app, result, states
from import_export.formats import base_formats
from picklefield import PickledObjectField

from . import tools
from .core import TaskStateInfo, TimeStampedModel


class ExportJob(TimeStampedModel):
    """Abstract model for managing celery export jobs.

    Encapsulate all logic related to celery export.

    Export steps:

    1. Create ExportJob with resource initialization parameters.
    2. Try to export all data to file.
    3. If everything correct - export data to file from database.

    Export file saves in media files.

    """

    class ExportStatus(models.TextChoices):
        """ExportJob possible statuses.

        * CREATED:
            export job is just created, no exporting done
        * EXPORTING:
            export job started
        * EXPORT_ERROR:
            DB queryset not exported, errors
        * EXPORTED:
            DB queryset exported, no errors occurred

        State diagrams::

            CREATED
               |
            EXPORTING  - (EXPORT_ERROR)
               |
            EXPORTED

        """

        CREATED = "CREATED", _("Created")
        EXPORTING = "EXPORTING", _("Exporting")
        EXPORT_ERROR = "EXPORT_ERROR", _("Export Error")
        EXPORTED = "EXPORTED", _("Exported")
        CANCELLED = "CANCELLED", _("Cancelled")

    export_finished_statuses = (
        ExportStatus.EXPORTED,
        ExportStatus.EXPORT_ERROR,
    )

    export_status = models.CharField(
        max_length=20,
        choices=ExportStatus.choices,
        default=ExportStatus.CREATED,
        verbose_name=_("Job status"),
    )

    resource_path = models.CharField(
        max_length=128,
        verbose_name=_("Resource class path"),
        help_text=_(
            "Dotted path to subclass of `import_export.Resource` that "
            "should be used for export",
        ),
    )

    file_format_path = models.CharField(
        max_length=128,
        verbose_name=_("Export path to class"),
        help_text=_(
            "Export file format as path to base_formats class",
        ),
    )

    data_file = models.FileField(
        max_length=512,
        verbose_name=_("Data file"),
        upload_to=tools.upload_export_file_to,
        help_text=_("File that contain exported data"),
    )

    resource_kwargs = models.JSONField(
        default=dict,
        verbose_name=_("Resource kwargs"),
        help_text=_("Keyword parameters required for resource initialization"),
    )

    traceback = models.TextField(
        blank=True,
        default=str,
        verbose_name=_("Traceback"),
        help_text=_("Python traceback in case of export error"),
    )
    error_message = models.CharField(
        max_length=128,
        blank=True,
        default=str,
        verbose_name=_("Error message"),
        help_text=_("Python error message in case of export error"),
    )

    result = PickledObjectField(
        default=str,
        verbose_name=_("Export result"),
        help_text=_(
            "Internal export result object that contain "
            "info about export statistics. Pickled Python object",
        ),
    )

    export_task_id = models.CharField(  # noqa: DJ01
        verbose_name=_("Export task ID"),
        max_length=36,
        null=True,
        blank=True,
        help_text=_("Celery task ID that start `export_data`"),
    )

    export_started = models.DateTimeField(
        verbose_name=_("Export started"),
        editable=False,
        blank=True,
        null=True,
    )

    export_finished = models.DateTimeField(
        verbose_name=_("Export finished"),
        editable=False,
        blank=True,
        null=True,
    )

    created_by = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Created by"),
        help_text=_("User which started export"),
    )

    class Meta:
        verbose_name = _("Export job")
        verbose_name_plural = _("Export jobs")

    def __str__(self) -> str:
        """Return string representation."""
        resource_name = pathlib.Path(self.resource_path).suffix.lstrip(".")
        file_format = pathlib.Path(self.file_format_path).suffix.lstrip(".")

        return f"ExportJob(resource={resource_name}, file_format={file_format})"

    def save(
        self,
        force_insert=False,
        force_update=False,
        using=None,
        update_fields=None,
    ):
        """Start task for data exporting when ExportJob is created.

        Celery task is manually called with `apply_async`, to provide
        possibility of custom `task_id` with which task will be run.

        """
        is_created = self._state.adding
        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )
        if is_created:
            self.export_task_id = str(uuid.uuid4())
            self.save(update_fields=["export_task_id"])
            transaction.on_commit(self._start_export_data_task)

    @property
    def resource(self):
        """Get initialized resource instance."""
        resource_class = module_loading.import_string(self.resource_path)
        resource = resource_class(
            created_by=self.created_by,
            **self.resource_kwargs,
        )
        return resource

    @property
    def file_format(self) -> base_formats.Format:
        """Get initialized format instance."""
        return module_loading.import_string(self.file_format_path)()

    @property
    def export_filename(self) -> str:
        """Get filename for export result file."""
        return self.resource.generate_export_filename(
            file_format=self.file_format,
        ).replace("/", "-")

    @property
    def progress(self) -> typing.Optional[TaskStateInfo]:
        """Return dict with parsing state.

        Example for sync mode::

            {
                'state': 'EXPORTING',
                'info': None
            }

        Example for celery (celery) mode::

            {
                'state': 'EXPORTING',
                'info': {'current': 15, 'total': 100}
            }

        Possible states:
            1. PENDING
            2. STARTED
            3. SUCCESS
            4. EXPORTING - custom status that also set export info

        https://docs.celeryproject.org/en/latest/userguide/tasks.html#states

        """
        if self.export_status not in (self.ExportStatus.EXPORTING,):
            return None

        if not self.export_task_id or current_app.conf.task_always_eager:
            return dict(
                state=self.export_status.upper(),
                info=None,
            )

        return self._get_task_state(self.export_task_id)

    def _check_import_status_correctness(
        self,
        expected_statuses: typing.Sequence[ExportStatus],
    ) -> None:
        """Raise `ValueError` if `ImportJob` is in incorrect state."""
        if self.export_status not in expected_statuses:
            raise ValueError(
                f"ImportJob with id {self.id} has incorrect status: "
                f"`{self.export_status}`. Expected statuses:"
                f" {[status.value for status in expected_statuses]}",
            )

    def _start_export_data_task(self):
        """Start export data task."""
        from .. import tasks

        tasks.export_data_task.apply_async(
            kwargs=dict(job_id=self.pk),
            task_id=self.export_task_id,
        )

    def export_data(self):
        """Export data to `data_file` from DB."""
        self.export_status = self.ExportStatus.EXPORTING
        self.export_started = timezone.now()
        self.save(
            update_fields=[
                "export_status",
                "export_started",
            ],
        )
        try:
            self._export_data_inner()
            self.export_status = self.ExportStatus.EXPORTED
            self.export_finished = timezone.now()
            self.save(
                update_fields=[
                    "export_status",
                    "export_finished",
                ],
            )
        except Exception as error:
            self.traceback = traceback.format_exc()
            self.error_message = str(error)[:512]
            self.export_status = self.ExportStatus.EXPORT_ERROR
            self.save(
                update_fields=[
                    "export_status",
                    "traceback",
                    "error_message",
                ],
            )

    def cancel_export(self) -> None:
        """Cancel current data export.

        ExportJob can be CANCELLED only from following states:
            - CREATED
            - EXPORTING

        """
        self._check_import_status_correctness(
            expected_statuses=[
                self.ExportStatus.CREATED.value,
                self.ExportStatus.EXPORTING.value,
            ],
        )

        # send signal to celery to revoke task
        current_app.control.revoke(self.export_task_id, terminate=True)

        self.export_status = self.ExportStatus.CANCELLED
        self.save(update_fields=["export_status"])

    def _export_data_inner(self):
        """Run export process with saving to file."""
        self.result = self.resource.export()
        self.save(update_fields=["result"])

        # `export_data` may be bytes (base formats such as xlsx, csv, etc.) or
        # file object (formats inherited from `BaseZipExport`)
        export_data = self.file_format.export_data(
            dataset=self.result,
            escape_output=self.resource_kwargs.get("escape_output", False),
            escape_html=self.resource_kwargs.get("escape_html", False),
            escape_formulae=self.resource_kwargs.get("escape_formulae", False),
        )
        # create file if `export_data` is not file
        if not hasattr(export_data, "read"):
            export_data = django_files.base.ContentFile(
                content=encoding.force_bytes(export_data),
            )
        self.data_file.save(
            name=self.export_filename,
            content=export_data,
            save=True,
        )

    def _get_task_state(self, task_id: str) -> TaskStateInfo:
        """Get state info for passed task_id.

        This method may change job status if task failed, but we did not
        save info about this to DB.

        This may happen if task start exporting, but processes was killed.
        In that case, job have status `importing`, but it can't be finished.

        """
        async_result = result.AsyncResult(task_id)
        if async_result.state not in states.EXCEPTION_STATES:
            return dict(
                state=async_result.state,
                info=async_result.info,
            )

        # Update job's status in case of exception
        self.export_status = ExportJob.ExportStatus.EXPORT_ERROR
        self.error_message = str(async_result.info)[:128]
        self.traceback = str(async_result.traceback)
        self.save(
            update_fields=[
                "error_message",
                "traceback",
                "export_status",
            ],
        )
        return dict(
            state=async_result.state,
            info={},
        )



from django import forms
from django.urls import reverse

from ... import models
from ..widgets import ProgressBarWidget


class ImportJobAdminForm(forms.ModelForm):
    """Admin form for ``ImportJob`` model.

    Adds custom `import_progressbar` field that displays current import
    progress using AJAX requests to specified endpoint. Fields widget is
    defined in `__init__` method.

    """

    import_progressbar = forms.Field(
        required=False,
    )

    def __init__(
        self,
        instance: models.ImportJob,
        *args,
        **kwargs,
    ):
        """Provide `import_progressbar` widget the ``ImportJob`` instance."""
        super().__init__(*args, instance=instance, **kwargs)
        url_name = "admin:import_job_progress"
        self.fields["import_progressbar"].label = (
            "Import progress" if
            instance.import_status == models.ImportJob.ImportStatus.IMPORTING
            else "Parsing progress"
        )
        self.fields["import_progressbar"].widget = ProgressBarWidget(
            job=instance,
            url=reverse(url_name, args=(instance.id,)),
        )

    class Meta:
        fields = (
            "import_status",
            "resource_path",
            "data_file",
            "resource_kwargs",
            "traceback",
            "error_message",
            "result",
            "parse_task_id",
            "import_task_id",
            "parse_finished",
            "import_started",
            "import_finished",
            "created_by",
            "created",
            "modified",
        )



from django.db.models.fields.reverse_related import ManyToManyRel

from import_export.fields import Field


class M2MField(Field):
    """Base M2M field provides faster related instances import."""

    def __init__(self, *args, export_order=None, **kwargs):
        """Save additional field params.

        Args:
            export_order(str): field name that should be used for ordering
                instances during export

        """
        super().__init__(*args, **kwargs)
        self.export_order = export_order

    def _format_exception(self, exception):
        """Shortcut for humanizing exception."""
        error = str(exception)
        if hasattr(exception, "messages"):
            error = " ".join(exception.messages)

        msg = f"Column '{self.column_name}': {error}"
        raise ValueError(msg) from exception

    def get_value(self, obj):
        """Return the value of the object's attribute.

        This method should return instances of intermediate model, i.e.
        Membership instances

        """
        if self.attribute is None:
            return None

        m2m_rel, _, field_name, _ = self.get_relation_field_params(obj)

        # retrieve intermediate model
        # intermediate_model == Membership
        intermediate_model = m2m_rel.through

        # filter relations with passed object
        qs = intermediate_model.objects.filter(**{field_name: obj})
        if self.export_order:
            qs = qs.order_by(self.export_order)
        return qs

    def get_relation_field_params(self, obj):
        """Shortcut to get relation field params.

        Gets relation, field itself, its name and its reversed field name

        """
        # retrieve M2M field itself (i.e. Artist.bands)
        field = obj._meta.get_field(self.attribute)

        # if field is `ManyToManyRel` - it is a reversed relation
        if isinstance(field, ManyToManyRel):
            m2m_rel = field
            m2m_field = m2m_rel.field
            field_name = m2m_field.m2m_reverse_field_name()
            reversed_field_name = m2m_field.m2m_field_name()
            return m2m_rel, m2m_field, field_name, reversed_field_name

        # otherwise it is a forward relation
        m2m_rel = field.remote_field
        m2m_field = field
        field_name = m2m_field.m2m_field_name()
        reversed_field_name = m2m_field.m2m_reverse_field_name()
        return m2m_rel, m2m_field, field_name, reversed_field_name

    def save(self, obj, data, *args, **kwargs):
        """Delete intermediate models.

        This implementation deletes intermediate models, which were excluded
        and creates intermediate models only for newly added models.

        Parent `save` method deletes and recreates intermediate models for all
        instances which generates a lot of exceed Feed Entries, so overridden.

        """
        (
            _,
            m2m_field,
            field_name,
            reversed_field_name,
        ) = self.get_relation_field_params(obj)

        # retrieve intermediate model (AttendeeTeamMembership)
        intermediate_model = m2m_field.remote_field.through

        # should be returned following list:
        # [{'object': Instance01, 'properties': {}},
        # {'object': Instance02, 'properties': {}}]
        data = self.clean(data)

        # IDs of related instances in imported data
        imported_ids = {i["object"].id for i in data}

        # IDs of current instances
        manager = getattr(obj, self.attribute)
        current_ids = set(manager.values_list("id", flat=True))

        # Find instances to be excluded after import
        excluded_ids = current_ids - imported_ids
        intermediate_model.objects.filter(
            **{
                field_name: obj,
                f"{reversed_field_name}__id__in": excluded_ids,
            },
        ).delete()

        # Find instances to add after import
        added_ids = imported_ids - current_ids
        for instance in data:
            # process only newly added attendees
            if instance["object"].id not in added_ids:
                continue
            obj_data = instance["properties"].copy()
            obj_data.update(
                {
                    field_name: obj,
                    reversed_field_name: instance["object"],
                },
            )
            intermediate_obj = intermediate_model(**obj_data)
            try:
                intermediate_obj.full_clean()
            except Exception as exception:
                self._format_exception(exception)
            intermediate_obj.save()


class IntermediateManyToManyField(M2MField):
    """Resource field for M2M with custom ``through`` model.

    By default, ``django-import-export`` set up object attributes using
    ``setattr(obj, attribute_name, value)``, where ``value`` is ``QuerySet``
    of related model objects. But django forbid this when `ManyToManyField``
    used with custom ``through`` model.

    This field expects be used with custom widget that return not simple value,
    but dict with intermediate model attributes.

    For easy comments following models will be used:

        Artist:
            name
            bands ->

        Membership:
            artist
            band
            date_joined

        Band:
            title
            <- artists

    So this field should be used for exporting Artists with `bands` field.

    Save workflow is following:
        1. clean data (extract dicts)
        2. Remove current M2M instances of object
        3. Create new M2M instances based on current object

    """

    def save(self, obj, data, *args, **kwargs):
        """Add M2M relations for obj from data.

        Args:
            obj(model instance): object being imported
            data(OrderedDict): all extracted data for object

        Example:
            obj - Artist instance

        """
        if self.readonly:
            return

        (
            m2m_rel,
            m2m_field,
            field_name,
            reversed_field_name,
        ) = self.get_relation_field_params(obj)

        # retrieve intermediate model
        # IntermediateModel == Membership
        IntermediateModel = m2m_rel.through

        # should be returned following list:
        # [{'band': <Band obj>, 'date_joined': '2016-08-18'}]
        instances_data = self.clean(data)

        # remove current related objects,
        # i.e. clear artists's band
        IntermediateModel.objects.filter(**{field_name: obj}).delete()

        for rel_obj_data in instances_data:
            # add current and remote object to intermediate instance data
            # i.e. {'artist': <Artist obj>, 'band': rel_obj_data['properties']}
            obj_data = rel_obj_data["properties"].copy()
            obj_data.update(
                {
                    field_name: obj,
                    reversed_field_name: rel_obj_data["object"],
                },
            )
            intermediate_obj = IntermediateModel(**obj_data)
            try:
                intermediate_obj.full_clean()
            except Exception as e:
                self._format_exception(e)
            intermediate_obj.save()
